# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import json

import pytest

from azure.ai.ml._restclient.v2024_01_01_preview.models import BatchDeployment as RestBatchDeployment
from azure.ai.ml._restclient.v2024_01_01_preview.models import (
    BatchDeploymentProperties as RestBatchDeploymentProperties,
)
from azure.ai.ml._restclient.v2024_01_01_preview.models import (
    BatchPipelineComponentDeploymentConfiguration as RestBatchPipelineComponentDeploymentConfiguration,
)
from azure.ai.ml._restclient.v2024_01_01_preview.models import IdAssetReference
from azure.ai.ml.entities import PipelineComponent
from azure.ai.ml.entities._deployment.pipeline_component_batch_deployment import PipelineComponentBatchDeployment
from azure.ai.ml.entities._load_functions import load_pipeline_component_batch_deployment


@pytest.mark.unittest
class TestPipelineComponentBatchDeployment:
    HELLO_BATCH_DEPLOYMENT = "tests/test_configs/deployments/batch/pipeline_component_batch_deployment.yml"
    HELLO_BATCH_DEPLOYMENT_REST = "tests/test_configs/deployments/batch/batch_pipeline_component_rest.json"

    def test_to_rest_object(self) -> None:
        pipeline_component = load_pipeline_component_batch_deployment(
            TestPipelineComponentBatchDeployment.HELLO_BATCH_DEPLOYMENT
        )
        assert pipeline_component.type == "pipeline"
        pipeline_component_rest = pipeline_component._to_rest_object(location="eastus")
        assert pipeline_component_rest.properties.deployment_configuration.settings == pipeline_component.settings
        assert (
            pipeline_component_rest.properties.deployment_configuration.component_id.asset_id
            == pipeline_component.component
        )
        assert pipeline_component_rest.properties.description == pipeline_component.description
        assert pipeline_component_rest.tags == pipeline_component.tags
        assert pipeline_component_rest.location == "eastus"

    def test_to_rest_object_for_component_obj(self) -> None:
        pipeline_component = load_pipeline_component_batch_deployment(
            TestPipelineComponentBatchDeployment.HELLO_BATCH_DEPLOYMENT
        )

        pipeline_component.component = PipelineComponent(
            id=pipeline_component.component,
            name="test_component",
            description="component description",
            tags={"tag1": "tag_value"},
        )
        pipeline_component_rest = pipeline_component._to_rest_object(location="eastus")

        assert pipeline_component_rest.properties.deployment_configuration.settings == pipeline_component.settings
        assert (
            pipeline_component_rest.properties.deployment_configuration.component_id.asset_id
            == pipeline_component.component.id
        )
        assert pipeline_component_rest.properties.deployment_configuration.tags == pipeline_component.component.tags
        assert (
            pipeline_component_rest.properties.deployment_configuration.description
            == pipeline_component.component.description
        )
        assert pipeline_component_rest.properties.description == pipeline_component.description
        assert pipeline_component_rest.tags == pipeline_component.tags
        assert pipeline_component_rest.location == "eastus"

    def test_to_dict(self) -> None:
        pipeline_component = load_pipeline_component_batch_deployment(
            TestPipelineComponentBatchDeployment.HELLO_BATCH_DEPLOYMENT,
            params_override=[{"endpoint_name": "azureml:hello_batch2@latest"}],
        )
        pipeline_component_dict = pipeline_component._to_dict()
        assert pipeline_component_dict["name"] == pipeline_component.name
        assert pipeline_component_dict["tags"] == pipeline_component.tags
        assert pipeline_component_dict["description"] == pipeline_component.description
        assert pipeline_component_dict["endpoint_name"] == "azureml:hello_batch2@latest"
        assert pipeline_component_dict["component"] == "azureml:hello_batch@latest"
        assert pipeline_component_dict["settings"] == pipeline_component.settings

    def test_from_rest_object(self) -> None:

        with open(TestPipelineComponentBatchDeployment.HELLO_BATCH_DEPLOYMENT_REST, "r") as file:
            pipeline_component_rest = RestBatchDeployment.from_dict(json.load(file))
            pipeline_component_rest.properties.additional_properties = {
                "deploymentConfiguration": {
                    "componentId": {"assetId": "azureml:hello_batch@latest"},
                    "settings": {"componentId": "azureml:hello_batch@latest"},
                }
            }
            pipeline_component_from_rest = PipelineComponentBatchDeployment._from_rest_object(pipeline_component_rest)
            assert pipeline_component_from_rest.component == "azureml:hello_batch@latest"
            assert (
                pipeline_component_from_rest.settings["componentId"]
                == pipeline_component_rest.properties.additional_properties["deploymentConfiguration"]["componentId"][
                    "assetId"
                ]
            )
            assert (
                pipeline_component_from_rest.component
                == pipeline_component_rest.properties.additional_properties["deploymentConfiguration"]["componentId"][
                    "assetId"
                ]
            )
            assert pipeline_component_from_rest.endpoint_name == "achauhan-endpoint-name"



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from azure.ai.ml.entities import Job
from azure.ai.ml.entities._job.base_job import _BaseJob
from azure.ai.ml.entities._system_data import SystemData

from ._restclient.index_service_apis import IndexServiceAPIs
from ._restclient.index_service_apis.models import (
    CrossRegionIndexEntitiesRequest,
    IndexEntitiesRequest,
    IndexEntitiesRequestFilter,
    IndexEntitiesRequestOrder,
    IndexEntitiesRequestOrderDirection,
    IndexEntityResponse,
)

__path__ = __import__("pkgutil").extend_path(__path__, __name__)  # type: ignore


def index_entity_response_to_job(entity: IndexEntityResponse) -> Job:
    properties = entity.properties
    annotations = entity.annotations
    creation_context = SystemData(
        created_by=properties.creation_context.created_by.user_name,
        created_by_type="User",
        created_at=properties.creation_context.created_time,
    )

    return _BaseJob(
        name=properties.additional_properties["runId"],
        display_name=annotations.additional_properties["displayName"],
        description=annotations.additional_properties["description"] or "",
        tags=annotations.tags,
        properties=properties.additional_properties["userProperties"],
        experiment_name=properties.additional_properties["experimentName"],
        services={},
        status=annotations.additional_properties["status"],
        creation_context=creation_context,
        compute=(
            properties.additional_properties["compute"]["armId"]
            if "compute" in properties.additional_properties
            else None
        ),
    )


__all__ = [
    "IndexServiceAPIs",
    "CrossRegionIndexEntitiesRequest",
    "IndexEntitiesRequest",
    "IndexEntitiesRequestFilter",
    "IndexEntitiesRequestOrder",
    "IndexEntitiesRequestOrderDirection",
    "index_entity_response_to_job",
]



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

import logging

from marshmallow import fields, post_load, pre_dump

from azure.ai.ml._schema.core.fields import ExperimentalField, NestedField
from azure.ai.ml._schema.core.intellectual_property import IntellectualPropertySchema
from azure.ai.ml._schema.core.schema import PathAwareSchema
from azure.ai.ml._schema.job import CreationContextSchema
from azure.ai.ml.constants._common import BASE_PATH_CONTEXT_KEY, AssetTypes, AzureMLResourceType

from ..core.fields import ArmVersionedStr, StringTransformedEnum, VersionField

module_logger = logging.getLogger(__name__)


class ModelSchema(PathAwareSchema):
    name = fields.Str(required=True)
    id = ArmVersionedStr(azureml_type=AzureMLResourceType.MODEL, dump_only=True)
    type = StringTransformedEnum(
        allowed_values=[
            AssetTypes.CUSTOM_MODEL,
            AssetTypes.MLFLOW_MODEL,
            AssetTypes.TRITON_MODEL,
        ],
        metadata={"description": "The storage format for this entity. Used for NCD."},
    )
    path = fields.Str()
    version = VersionField()
    description = fields.Str()
    properties = fields.Dict()
    tags = fields.Dict()
    stage = fields.Str()
    utc_time_created = fields.DateTime(format="iso", dump_only=True)
    flavors = fields.Dict()
    creation_context = NestedField(CreationContextSchema, dump_only=True)
    job_name = fields.Str(dump_only=True)
    latest_version = fields.Str(dump_only=True)
    datastore = fields.Str(metadata={"description": "Name of the datastore to upload to."}, required=False)
    intellectual_property = ExperimentalField(NestedField(IntellectualPropertySchema, required=False), dump_only=True)
    system_metadata = fields.Dict()

    @pre_dump
    def validate(self, data, **kwargs):
        if data._intellectual_property:  # pylint: disable=protected-access
            ipp_field = data._intellectual_property  # pylint: disable=protected-access
            if ipp_field:
                setattr(data, "intellectual_property", ipp_field)
        return data

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._assets import Model

        return Model(base_path=self.context[BASE_PATH_CONTEXT_KEY], **data)


class AnonymousModelSchema(ModelSchema):
    name = fields.Str()
    version = VersionField()



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

import warnings
from os import PathLike
from pathlib import Path
from typing import IO, Any, AnyStr, Dict, List, Optional, Type, Union, cast


from azure.ai.ml._restclient.v2024_04_01_preview.models import (
    WorkspaceConnectionPropertiesV2BasicResource as RestWorkspaceConnection,
)
from azure.ai.ml._restclient.v2024_04_01_preview.models import (
    ConnectionCategory,
    NoneAuthTypeWorkspaceConnectionProperties,
    AADAuthTypeWorkspaceConnectionProperties,
)

from azure.ai.ml._schema.workspace.connections.workspace_connection import WorkspaceConnectionSchema
from azure.ai.ml._utils.utils import _snake_to_camel, camel_to_snake, dump_yaml_to_file
from azure.ai.ml.constants._common import (
    BASE_PATH_CONTEXT_KEY,
    PARAMS_OVERRIDE_KEY,
    ConnectionTypes,
    CognitiveServiceKinds,
    CONNECTION_KIND_KEY,
    CONNECTION_RESOURCE_ID_KEY,
)
from azure.ai.ml.entities._credentials import (
    AccessKeyConfiguration,
    ApiKeyConfiguration,
    ManagedIdentityConfiguration,
    NoneCredentialConfiguration,
    PatTokenConfiguration,
    SasTokenConfiguration,
    ServicePrincipalConfiguration,
    UsernamePasswordConfiguration,
    _BaseIdentityConfiguration,
    AccountKeyConfiguration,
    AadCredentialConfiguration,
)
from azure.ai.ml.entities._resource import Resource
from azure.ai.ml.entities._system_data import SystemData
from azure.ai.ml.entities._util import load_from_dict


CONNECTION_CATEGORY_TO_CREDENTIAL_MAP = {
    ConnectionCategory.AZURE_BLOB: [AccessKeyConfiguration, SasTokenConfiguration, AadCredentialConfiguration],
    ConnectionTypes.AZURE_DATA_LAKE_GEN_2: [
        ServicePrincipalConfiguration,
        AadCredentialConfiguration,
        ManagedIdentityConfiguration,
    ],
    ConnectionCategory.GIT: [PatTokenConfiguration, NoneCredentialConfiguration, UsernamePasswordConfiguration],
    ConnectionCategory.PYTHON_FEED: [UsernamePasswordConfiguration, PatTokenConfiguration, NoneCredentialConfiguration],
    ConnectionCategory.CONTAINER_REGISTRY: [ManagedIdentityConfiguration, UsernamePasswordConfiguration],
}

DATASTORE_CONNECTIONS = {
    ConnectionCategory.AZURE_BLOB,
    ConnectionTypes.AZURE_DATA_LAKE_GEN_2,
    ConnectionCategory.AZURE_ONE_LAKE,
}

CONNECTION_ALTERNATE_TARGET_NAMES = ["target", "api_base", "url", "azure_endpoint", "endpoint"]


# Dev note: The acceptable strings for the type field are all snake_cased versions of the string constants defined
# In the rest client enum defined at _azure_machine_learning_services_enums.ConnectionCategory.
# We avoid directly referencing it in the docs to avoid restclient references.
class WorkspaceConnection(Resource):
    """Azure ML connection provides a secure way to store authentication and configuration information needed
    to connect and interact with the external resources.

    Note: For connections to OpenAI, Cognitive Search, and Cognitive Services, use the respective subclasses
    (ex: ~azure.ai.ml.entities.OpenAIConnection) instead of instantiating this class directly.

    :param name: Name of the connection.
    :type name: str
    :param target: The URL or ARM resource ID of the external resource.
    :type target: str
    :param metadata: Metadata dictionary.
    :type metadata: Optional[Dict[str, Any]]
    :param type: The category of external resource for this connection.
    :type type: The type of connection, possible values are: "git", "python_feed", "container_registry",
        "feature_store", "s3", "snowflake", "azure_sql_db", "azure_synapse_analytics", "azure_my_sql_db",
        "azure_postgres_db", "adls_gen_2", "azure_one_lake", "custom".
    :param credentials: The credentials for authenticating to the external resource. Note that certain connection
        types (as defined by the type input) only accept certain types of credentials.
    :type credentials: Union[
        ~azure.ai.ml.entities.PatTokenConfiguration,
        ~azure.ai.ml.entities.SasTokenConfiguration,
        ~azure.ai.ml.entities.UsernamePasswordConfiguration,
        ~azure.ai.ml.entities.ManagedIdentityConfiguration
        ~azure.ai.ml.entities.ServicePrincipalConfiguration,
        ~azure.ai.ml.entities.AccessKeyConfiguration,
        ~azure.ai.ml.entities.ApiKeyConfiguration,
        ~azure.ai.ml.entities.NoneCredentialConfiguration
        ~azure.ai.ml.entities.AccountKeyConfiguration,
        ~azure.ai.ml.entities.AadCredentialConfiguration,
        None
        ]
    :param is_shared: For connections in project, this controls whether or not this connection
        is shared amongst other projects that are shared by the parent hub. Defaults to true.
    :type is_shared: bool
    """

    def __init__(
        self,
        *,
        # TODO : Check if this is okay since it shadows builtin-type type
        type: str,  # pylint: disable=redefined-builtin
        credentials: Union[
            PatTokenConfiguration,
            SasTokenConfiguration,
            UsernamePasswordConfiguration,
            ManagedIdentityConfiguration,
            ServicePrincipalConfiguration,
            AccessKeyConfiguration,
            ApiKeyConfiguration,
            NoneCredentialConfiguration,
            AccountKeyConfiguration,
            AadCredentialConfiguration,
        ],
        is_shared: bool = True,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):

        # Dev note: This initializer has an undocumented kwarg "from_child" to determine if this initialization
        # is from a child class.
        # This kwarg is required to allow instantiation of types that are associated with subtypes without a
        # warning printout.
        # The additional undocumented kwarg "strict_typing" turns the warning into a value error.
        from_child = kwargs.pop("from_child", False)
        strict_typing = kwargs.pop("strict_typing", False)
        correct_class = WorkspaceConnection._get_entity_class_from_type(type)
        if not from_child and correct_class != WorkspaceConnection:
            if strict_typing:
                raise ValueError(
                    f"Cannot instantiate a base Connection with a type of {type}. "
                    f"Please use the appropriate subclass {correct_class.__name__} instead."
                )
            warnings.warn(
                f"The connection of {type} has additional fields and should not be instantiated directly "
                f"from the Connection class. Please use its subclass {correct_class.__name__} instead.",
            )
        # This disgusting code allows for a variety of inputs names to technically all
        # act like the target field, while still maintaining the aggregate field as required.
        target = None
        for target_name in CONNECTION_ALTERNATE_TARGET_NAMES:
            target = kwargs.pop(target_name, target)
        if target is None and type not in {ConnectionCategory.SERP, ConnectionCategory.Open_AI}:
            raise ValueError("target is a required field for Connection.")

        tags = kwargs.pop("tags", None)
        if tags is not None:
            if metadata is not None:
                # Update tags updated with metadata to make sure metadata values are preserved in case of conflicts.
                tags.update(metadata)
                metadata = tags
                warnings.warn(
                    "Tags are a deprecated field for connections, use metadata instead. Since both "
                    + "metadata and tags are assigned, metadata values will take precedence in the event of conflicts."
                )
            else:
                metadata = tags
                warnings.warn("Tags are a deprecated field for connections, use metadata instead.")

        super().__init__(**kwargs)

        self.type = type
        self._target = target
        self._credentials = credentials
        self._is_shared = is_shared
        self._metadata = metadata
        self._validate_cred_for_conn_cat()

    def _validate_cred_for_conn_cat(self) -> None:
        """Given a connection type, ensure that the given credentials are valid for that connection type.
        Does not validate the actual data of the inputted credential, just that they are of the right class
        type.

        """
        # Convert none credentials to AAD credentials for datastore connection types.
        # The backend stores datastore aad creds as none, unlike other connection types with aad,
        # which actually list them as aad. This IS distinct from regular none credentials, or so I've been told,
        # so I will endeavor to smooth over that inconsistency here.
        converted_type = _snake_to_camel(self.type).lower()
        if self._credentials == NoneCredentialConfiguration() and any(
            converted_type == _snake_to_camel(item).lower() for item in DATASTORE_CONNECTIONS
        ):
            self._credentials = AadCredentialConfiguration()

        if self.type in CONNECTION_CATEGORY_TO_CREDENTIAL_MAP:
            allowed_credentials = CONNECTION_CATEGORY_TO_CREDENTIAL_MAP[self.type]
            if self.credentials is None and NoneCredentialConfiguration not in allowed_credentials:
                raise ValueError(
                    f"Cannot instantiate a Connection with a type of {self.type} and no credentials."
                    f"Please supply credentials from one of the following types: {allowed_credentials}."
                )
            cred_type = type(self.credentials)
            if cred_type not in allowed_credentials:
                raise ValueError(
                    f"Cannot instantiate a Connection with a type of {self.type} and credentials of type"
                    f" {cred_type}. Please supply credentials from one of the following types: {allowed_credentials}."
                )
        # For unknown types, just let the user do whatever they want.

    @property
    def type(self) -> Optional[str]:
        """Type of the connection, supported are 'git', 'python_feed' and 'container_registry'.

        :return: Type of the job.
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, value: str) -> None:
        """Set the type of the connection, supported are 'git', 'python_feed' and 'container_registry'.

        :param value: value for the type of connection.
        :type: str
        """
        if not value:
            return
        self._type: Optional[str] = camel_to_snake(value)

    @property
    def target(self) -> Optional[str]:
        """Target url for the connection.

        :return: Target of the connection.
        :rtype: Optional[str]
        """
        return self._target

    @property
    def endpoint(self) -> Optional[str]:
        """Alternate name for the target of the connection,
        which is used by some connection subclasses.

        :return: The target of the connection.
        :rtype: str
        """
        return self.target

    @property
    def azure_endpoint(self) -> Optional[str]:
        """Alternate name for the target of the connection,
        which is used by some connection subclasses.

        :return: The target of the connection.
        :rtype: str
        """
        return self.target

    @property
    def url(self) -> Optional[str]:
        """Alternate name for the target of the connection,
        which is used by some connection subclasses.

        :return: The target of the connection.
        :rtype: str
        """
        return self.target

    @property
    def api_base(self) -> Optional[str]:
        """Alternate name for the target of the connection,
        which is used by some connection subclasses.

        :return: The target of the connection.
        :rtype: str
        """
        return self.target

    @property
    def credentials(
        self,
    ) -> Union[
        PatTokenConfiguration,
        SasTokenConfiguration,
        UsernamePasswordConfiguration,
        ManagedIdentityConfiguration,
        ServicePrincipalConfiguration,
        AccessKeyConfiguration,
        ApiKeyConfiguration,
        NoneCredentialConfiguration,
        AccountKeyConfiguration,
        AadCredentialConfiguration,
    ]:
        """Credentials for connection.

        :return: Credentials for connection.
        :rtype: Union[
            ~azure.ai.ml.entities.PatTokenConfiguration,
            ~azure.ai.ml.entities.SasTokenConfiguration,
            ~azure.ai.ml.entities.UsernamePasswordConfiguration,
            ~azure.ai.ml.entities.ManagedIdentityConfiguration
            ~azure.ai.ml.entities.ServicePrincipalConfiguration,
            ~azure.ai.ml.entities.AccessKeyConfiguration,
            ~azure.ai.ml.entities.ApiKeyConfiguration
            ~azure.ai.ml.entities.NoneCredentialConfiguration,
            ~azure.ai.ml.entities.AccountKeyConfiguration,
            ~azure.ai.ml.entities.AadCredentialConfiguration,
            ]
        """
        return self._credentials

    @property
    def metadata(self) -> Optional[Dict[str, Any]]:
        """The connection's metadata dictionary.
        :return: This connection's metadata.
        :rtype: Optional[Dict[str, Any]]
        """
        return self._metadata if self._metadata is not None else {}

    @metadata.setter
    def metadata(self, value: Optional[Dict[str, Any]]) -> None:
        """Set the metadata for the connection. Be warned that setting this will override
        ALL metadata values, including those implicitly set by certain connection types to manage their
        extra data. Usually, you should probably access the metadata dictionary, then add or remove values
        individually as needed.
        :param value: The new metadata for connection.
            This completely overwrites the existing metadata dictionary.
        :type value: Optional[Dict[str, Any]]
        """
        if not value:
            return
        self._metadata = value

    @property
    def tags(self) -> Optional[Dict[str, Any]]:
        """Deprecated. Use metadata instead.
        :return: This connection's metadata.
        :rtype: Optional[Dict[str, Any]]
        """
        return self._metadata if self._metadata is not None else {}

    @tags.setter
    def tags(self, value: Optional[Dict[str, Any]]) -> None:
        """Deprecated use metadata instead
        :param value: The new metadata for connection.
            This completely overwrites the existing metadata dictionary.
        :type value: Optional[Dict[str, Any]]
        """
        if not value:
            return
        self._metadata = value

    @property
    def is_shared(self) -> bool:
        """Get the Boolean describing if this connection is shared amongst its cohort within a hub.
        Only applicable for connections created within a project.

        :rtype: bool
        """
        return self._is_shared

    @is_shared.setter
    def is_shared(self, value: bool) -> None:
        """Assign the is_shared property of the connection, determining if it is shared amongst other projects
        within its parent hub. Only applicable for connections created within a project.

        :param value: The new is_shared value.
        :type value: bool
        """
        if not value:
            return
        self._is_shared = value

    def dump(self, dest: Union[str, PathLike, IO[AnyStr]], **kwargs: Any) -> None:
        """Dump the connection spec into a file in yaml format.

        :param dest: The destination to receive this connection's spec.
            Must be either a path to a local file, or an already-open file stream.
            If dest is a file path, a new file will be created,
            and an exception is raised if the file exists.
            If dest is an open file, the file will be written to directly,
            and an exception will be raised if the file is not writable.
        :type dest: Union[PathLike, str, IO[AnyStr]]
        """
        path = kwargs.pop("path", None)
        yaml_serialized = self._to_dict()
        dump_yaml_to_file(dest, yaml_serialized, default_flow_style=False, path=path, **kwargs)

    @classmethod
    def _load(
        cls,
        data: Optional[Dict] = None,
        yaml_path: Optional[Union[PathLike, str]] = None,
        params_override: Optional[list] = None,
        **kwargs: Any,
    ) -> "WorkspaceConnection":
        data = data or {}
        params_override = params_override or []
        context = {
            BASE_PATH_CONTEXT_KEY: Path(yaml_path).parent if yaml_path else Path("./"),
            PARAMS_OVERRIDE_KEY: params_override,
        }
        return cls._load_from_dict(data=data, context=context, **kwargs)

    @classmethod
    def _load_from_dict(cls, data: Dict, context: Dict, **kwargs: Any) -> "WorkspaceConnection":
        conn_type = data["type"] if "type" in data else None
        schema_class = cls._get_schema_class_from_type(conn_type)
        loaded_data: WorkspaceConnection = load_from_dict(schema_class, data, context, **kwargs)
        return loaded_data

    def _to_dict(self) -> Dict:
        schema_class = WorkspaceConnection._get_schema_class_from_type(self.type)
        # Not sure what this pylint complaint was about, probably due to the polymorphic
        # tricks at play. Disabling since testing indicates no issue.
        res: dict = schema_class(context={BASE_PATH_CONTEXT_KEY: "./"}).dump(self)
        return res

    @classmethod
    def _from_rest_object(cls, rest_obj: RestWorkspaceConnection) -> "WorkspaceConnection":
        conn_class = cls._get_entity_class_from_rest_obj(rest_obj)

        popped_metadata = conn_class._get_required_metadata_fields()

        rest_kwargs = cls._extract_kwargs_from_rest_obj(rest_obj=rest_obj, popped_metadata=popped_metadata)
        # Check for alternative name for custom connection type (added for client clarity).
        if rest_kwargs["type"].lower() == camel_to_snake(ConnectionCategory.CUSTOM_KEYS).lower():
            rest_kwargs["type"] = ConnectionTypes.CUSTOM
        if rest_kwargs["type"].lower() == camel_to_snake(ConnectionCategory.ADLS_GEN2).lower():
            rest_kwargs["type"] = ConnectionTypes.AZURE_DATA_LAKE_GEN_2
        target = rest_kwargs.get("target", "")
        # This dumb code accomplishes 2 things.
        # It ensures that sub-classes properly input their target, regardless of which
        # arbitrary name they replace it with, while also still allowing our official
        # client specs to list those inputs as 'required'
        for target_name in CONNECTION_ALTERNATE_TARGET_NAMES:
            rest_kwargs[target_name] = target
        if rest_obj.properties.category == ConnectionCategory.AZURE_ONE_LAKE:
            # The microsoft one lake connection uniquely has client-only inputs
            # that aren't just an alternate name for the target.
            # This sets those inputs, that way the initializer can still
            # required those fields for users.
            rest_kwargs["artifact"] = ""
            rest_kwargs["one_lake_workspace_name"] = ""
        if rest_obj.properties.category == ConnectionTypes.AI_SERVICES_REST_PLACEHOLDER:
            # AI Services renames it's metadata field when surfaced to users and inputted
            # into it's initializer for clarity. ResourceId doesn't really tell much on its own.
            # No default in pop, this should fail if we somehow don't get a resource ID
            rest_kwargs["ai_services_resource_id"] = rest_kwargs.pop(camel_to_snake(CONNECTION_RESOURCE_ID_KEY))
        connection = conn_class(**rest_kwargs)
        return cast(WorkspaceConnection, connection)

    def _validate(self) -> str:
        return str(self.name)

    def _to_rest_object(self) -> RestWorkspaceConnection:
        connection_properties_class: Any = NoneAuthTypeWorkspaceConnectionProperties
        if self._credentials:
            connection_properties_class = self._credentials._get_rest_properties_class()
        # Convert from human readable type to corresponding api enum if needed.
        conn_type = self.type
        if conn_type == ConnectionTypes.CUSTOM:
            conn_type = ConnectionCategory.CUSTOM_KEYS
        elif conn_type == ConnectionTypes.AZURE_DATA_LAKE_GEN_2:
            conn_type = ConnectionCategory.ADLS_GEN2
        elif conn_type in {
            ConnectionTypes.AZURE_CONTENT_SAFETY,
            ConnectionTypes.AZURE_SPEECH_SERVICES,
        }:
            conn_type = ConnectionCategory.COGNITIVE_SERVICE
        elif conn_type == ConnectionTypes.AZURE_SEARCH:
            conn_type = ConnectionCategory.COGNITIVE_SEARCH
        elif conn_type == ConnectionTypes.AZURE_AI_SERVICES:
            # ConnectionCategory.AI_SERVICES category accidentally unpublished
            conn_type = ConnectionTypes.AI_SERVICES_REST_PLACEHOLDER
        # Some credential property bags have no credential input.
        if connection_properties_class in {
            NoneAuthTypeWorkspaceConnectionProperties,
            AADAuthTypeWorkspaceConnectionProperties,
        }:
            properties = connection_properties_class(
                target=self.target,
                metadata=self.metadata,
                category=_snake_to_camel(conn_type),
                is_shared_to_all=self.is_shared,
            )
        else:
            properties = connection_properties_class(
                target=self.target,
                credentials=self.credentials._to_workspace_connection_rest_object() if self._credentials else None,
                metadata=self.metadata,
                category=_snake_to_camel(conn_type),
                is_shared_to_all=self.is_shared,
            )

        return RestWorkspaceConnection(properties=properties)

    @classmethod
    def _extract_kwargs_from_rest_obj(
        cls, rest_obj: RestWorkspaceConnection, popped_metadata: List[str]
    ) -> Dict[str, str]:
        """Internal helper function with extracts all the fields needed to initialize a connection object
        from its associated restful object. Pulls extra fields based on the supplied `popped_metadata` input.
        Returns all the fields as a dictionary, which is expected to then be supplied to a
        connection initializer as kwargs.

        :param rest_obj: The rest object representation of a connection
        :type rest_obj: RestWorkspaceConnection
        :param popped_metadata: Key names that should be pulled from the rest object's metadata and
            injected as top-level fields into the client connection's initializer.
            This is needed for subclasses that require extra inputs compared to the base Connection class.
        :type popped_metadata: List[str]

        :return: A dictionary containing all kwargs needed to construct a connection.
        :rtype: Dict[str, str]
        """
        properties = rest_obj.properties
        credentials: Any = NoneCredentialConfiguration()

        credentials_class = _BaseIdentityConfiguration._get_credential_class_from_rest_type(properties.auth_type)
        # None and AAD auth types have a property bag class, but no credentials inside that.
        # Thankfully they both have no inputs.

        if credentials_class is AadCredentialConfiguration:
            credentials = AadCredentialConfiguration()
        elif credentials_class is not NoneCredentialConfiguration:
            credentials = credentials_class._from_workspace_connection_rest_object(properties.credentials)

        metadata = properties.metadata if hasattr(properties, "metadata") else {}
        rest_kwargs = {
            "id": rest_obj.id,
            "name": rest_obj.name,
            "target": properties.target,
            "creation_context": SystemData._from_rest_object(rest_obj.system_data) if rest_obj.system_data else None,
            "type": camel_to_snake(properties.category),
            "credentials": credentials,
            "metadata": metadata,
            "is_shared": properties.is_shared_to_all if hasattr(properties, "is_shared_to_all") else True,
        }

        for name in popped_metadata:
            if name in metadata:
                rest_kwargs[camel_to_snake(name)] = metadata[name]
        return rest_kwargs

    @classmethod
    def _get_entity_class_from_type(cls, type: str) -> Type:
        """Helper function that derives the correct connection class given the client or server type.
        Differs slightly from the rest object version in that it doesn't need to account for
        rest object metadata.

        This reason there are two functions at all is due to certain API connection types that
        are obfuscated with different names when presented to the client. These types are
        accounted for in the ConnectionTypes class in the constants file.

        :param type: The type string describing the connection.
        :type type: str

        :return: Theconnection class the conn_type corresponds to.
        :rtype: Type
        """
        from .connection_subtypes import (
            AzureBlobStoreConnection,
            MicrosoftOneLakeConnection,
            AzureOpenAIConnection,
            AzureAIServicesConnection,
            AzureAISearchConnection,
            AzureContentSafetyConnection,
            AzureSpeechServicesConnection,
            APIKeyConnection,
            OpenAIConnection,
            SerpConnection,
            ServerlessConnection,
        )

        conn_type = _snake_to_camel(type).lower()
        if conn_type is None:
            return WorkspaceConnection

        # Connection categories don't perfectly follow perfect camel casing, so lower
        # case everything to avoid problems.
        CONNECTION_CATEGORY_TO_SUBCLASS_MAP = {
            ConnectionCategory.AZURE_OPEN_AI.lower(): AzureOpenAIConnection,
            ConnectionCategory.AZURE_BLOB.lower(): AzureBlobStoreConnection,
            ConnectionCategory.AZURE_ONE_LAKE.lower(): MicrosoftOneLakeConnection,
            ConnectionCategory.API_KEY.lower(): APIKeyConnection,
            ConnectionCategory.OPEN_AI.lower(): OpenAIConnection,
            ConnectionCategory.SERP.lower(): SerpConnection,
            ConnectionCategory.SERVERLESS.lower(): ServerlessConnection,
            _snake_to_camel(ConnectionTypes.AZURE_CONTENT_SAFETY).lower(): AzureContentSafetyConnection,
            _snake_to_camel(ConnectionTypes.AZURE_SPEECH_SERVICES).lower(): AzureSpeechServicesConnection,
            ConnectionCategory.COGNITIVE_SEARCH.lower(): AzureAISearchConnection,
            _snake_to_camel(ConnectionTypes.AZURE_SEARCH).lower(): AzureAISearchConnection,
            _snake_to_camel(ConnectionTypes.AZURE_AI_SERVICES).lower(): AzureAIServicesConnection,
            ConnectionTypes.AI_SERVICES_REST_PLACEHOLDER.lower(): AzureAIServicesConnection,
        }
        return CONNECTION_CATEGORY_TO_SUBCLASS_MAP.get(conn_type, WorkspaceConnection)

    @classmethod
    def _get_entity_class_from_rest_obj(cls, rest_obj: RestWorkspaceConnection) -> Type:
        """Helper function that converts a restful connection into the associated
         connection class or subclass. Accounts for potential snake/camel case and
        capitalization differences in the type, and sub-typing derived from metadata.

        :param rest_obj: The rest object representation of the connection to derive a class from.
        :type rest_obj: RestWorkspaceConnection

        :return: The  connection class the conn_type corresponds to.
        :rtype: Type
        """
        conn_type = rest_obj.properties.category
        conn_type = _snake_to_camel(conn_type).lower()
        if conn_type is None:
            return WorkspaceConnection

        # Imports are done here to avoid circular imports on load.
        from .connection_subtypes import (
            AzureContentSafetyConnection,
            AzureSpeechServicesConnection,
        )

        # Cognitive search connections have further subdivisions based on the kind of service.
        if (
            conn_type == ConnectionCategory.COGNITIVE_SERVICE.lower()
            and hasattr(rest_obj.properties, "metadata")
            and rest_obj.properties.metadata is not None
        ):
            kind = rest_obj.properties.metadata.get(CONNECTION_KIND_KEY, "").lower()
            if kind == CognitiveServiceKinds.CONTENT_SAFETY.lower():
                return AzureContentSafetyConnection
            if kind == CognitiveServiceKinds.SPEECH.lower():
                return AzureSpeechServicesConnection
            return WorkspaceConnection

        return cls._get_entity_class_from_type(type=conn_type)

    @classmethod
    def _get_schema_class_from_type(cls, conn_type: Optional[str]) -> Type:
        """Helper function that converts a rest client connection category into the associated
        connection schema class or subclass. Accounts for potential snake/camel case and
        capitalization differences.

        :param conn_type: The connection type.
        :type conn_type: str

        :return: The  connection schema class the conn_type corresponds to.
        :rtype: Type
        """
        if conn_type is None:
            return WorkspaceConnectionSchema
        entity_class = cls._get_entity_class_from_type(conn_type)
        return entity_class._get_schema_class()

    @classmethod
    def _get_required_metadata_fields(cls) -> List[str]:
        """Helper function that returns the required metadata fields for specific
        connection type. This parent function returns nothing, but needs to be overwritten by child
        classes, which are created under the expectation that they have extra fields that need to be
        accounted for.

        :return: A list of the required metadata fields for the specific connection type.
        :rtype: List[str]
        """
        return []

    @classmethod
    def _get_schema_class(cls) -> Type:
        """Helper function that maps this class to its associated schema class. Needs to be overridden by
        child classes to allow the base class to be polymorphic in its schema reading.

        :return: The appropriate schema class to use with this entity class.
        :rtype: Type
        """
        return WorkspaceConnectionSchema



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import abc
from typing import TYPE_CHECKING, Any, Optional, cast

from ..._exceptions import CredentialUnavailableError
from . import AsyncContextManager
from .get_token_mixin import GetTokenMixin
from .managed_identity_client import AsyncManagedIdentityClient

if TYPE_CHECKING:
    from azure.core.credentials import AccessToken


class AsyncManagedIdentityBase(AsyncContextManager, GetTokenMixin):
    """Base class for internal credentials using AsyncManagedIdentityClient."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        self._client = self.get_client(**kwargs)

    @abc.abstractmethod
    def get_client(self, **kwargs: Any) -> "Optional[AsyncManagedIdentityClient]":
        pass

    @abc.abstractmethod
    def get_unavailable_message(self) -> str:
        pass

    async def __aenter__(self) -> "AsyncManagedIdentityBase":
        if self._client:
            await self._client.__aenter__()
        return self

    async def __aexit__(self, *args: Any) -> None:
        if self._client:
            await self._client.__aexit__(*args)

    async def close(self) -> None:
        await self.__aexit__()

    async def get_token(self, *scopes: str, **kwargs: "Any") -> "AccessToken":
        if not self._client:
            raise CredentialUnavailableError(message=self.get_unavailable_message())
        return await super().get_token(*scopes, **kwargs)

    async def _acquire_token_silently(self, *scopes: str, **kwargs: "Any") -> "Optional[AccessToken]":
        # casting because mypy can't determine that these methods are called
        # only by get_token, which raises when self._client is None
        return cast(AsyncManagedIdentityClient, self._client).get_cached_token(*scopes)

    async def _request_token(self, *scopes: str, **kwargs: "Any") -> "AccessToken":
        return await cast(AsyncManagedIdentityClient, self._client).request_token(*scopes, **kwargs)



# pylint: disable=too-many-lines
# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
from typing import TYPE_CHECKING

from msrest import Serializer

from azure.core.exceptions import ClientAuthenticationError, HttpResponseError, ResourceExistsError, ResourceNotFoundError, map_error
from azure.core.paging import ItemPaged
from azure.core.pipeline import PipelineResponse
from azure.core.pipeline.transport import HttpResponse
from azure.core.rest import HttpRequest
from azure.core.tracing.decorator import distributed_trace
from azure.mgmt.core.exceptions import ARMErrorFormat

from .. import models as _models
from .._vendor import _convert_request, _format_url_section

if TYPE_CHECKING:
    # pylint: disable=unused-import,ungrouped-imports
    from typing import Any, Callable, Dict, Iterable, Optional, TypeVar, Union
    T = TypeVar('T')
    ClsType = Optional[Callable[[PipelineResponse[HttpRequest, HttpResponse], T, Dict[str, Any]], Any]]

_SERIALIZER = Serializer()
_SERIALIZER.client_side_validation = False
# fmt: off

def build_list_request(
    subscription_id,  # type: str
    resource_group_name,  # type: str
    workspace_name,  # type: str
    **kwargs  # type: Any
):
    # type: (...) -> HttpRequest
    api_version = kwargs.pop('api_version', "2024-04-01-preview")  # type: str
    skip = kwargs.pop('skip', None)  # type: Optional[str]
    list_view_type = kwargs.pop('list_view_type', None)  # type: Optional[Union[str, "_models.ListViewType"]]

    accept = "application/json"
    # Construct URL
    _url = kwargs.pop("template_url", "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/components")  # pylint: disable=line-too-long
    path_format_arguments = {
        "subscriptionId": _SERIALIZER.url("subscription_id", subscription_id, 'str', min_length=1),
        "resourceGroupName": _SERIALIZER.url("resource_group_name", resource_group_name, 'str', max_length=90, min_length=1),
        "workspaceName": _SERIALIZER.url("workspace_name", workspace_name, 'str', pattern=r'^[a-zA-Z0-9][a-zA-Z0-9_-]{2,32}$'),
    }

    _url = _format_url_section(_url, **path_format_arguments)

    # Construct parameters
    _query_parameters = kwargs.pop("params", {})  # type: Dict[str, Any]
    _query_parameters['api-version'] = _SERIALIZER.query("api_version", api_version, 'str')
    if skip is not None:
        _query_parameters['$skip'] = _SERIALIZER.query("skip", skip, 'str')
    if list_view_type is not None:
        _query_parameters['listViewType'] = _SERIALIZER.query("list_view_type", list_view_type, 'str')

    # Construct headers
    _header_parameters = kwargs.pop("headers", {})  # type: Dict[str, Any]
    _header_parameters['Accept'] = _SERIALIZER.header("accept", accept, 'str')

    return HttpRequest(
        method="GET",
        url=_url,
        params=_query_parameters,
        headers=_header_parameters,
        **kwargs
    )


def build_delete_request(
    subscription_id,  # type: str
    resource_group_name,  # type: str
    workspace_name,  # type: str
    name,  # type: str
    **kwargs  # type: Any
):
    # type: (...) -> HttpRequest
    api_version = kwargs.pop('api_version', "2024-04-01-preview")  # type: str

    accept = "application/json"
    # Construct URL
    _url = kwargs.pop("template_url", "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/components/{name}")  # pylint: disable=line-too-long
    path_format_arguments = {
        "subscriptionId": _SERIALIZER.url("subscription_id", subscription_id, 'str', min_length=1),
        "resourceGroupName": _SERIALIZER.url("resource_group_name", resource_group_name, 'str', max_length=90, min_length=1),
        "workspaceName": _SERIALIZER.url("workspace_name", workspace_name, 'str', pattern=r'^[a-zA-Z0-9][a-zA-Z0-9_-]{2,32}$'),
        "name": _SERIALIZER.url("name", name, 'str'),
    }

    _url = _format_url_section(_url, **path_format_arguments)

    # Construct parameters
    _query_parameters = kwargs.pop("params", {})  # type: Dict[str, Any]
    _query_parameters['api-version'] = _SERIALIZER.query("api_version", api_version, 'str')

    # Construct headers
    _header_parameters = kwargs.pop("headers", {})  # type: Dict[str, Any]
    _header_parameters['Accept'] = _SERIALIZER.header("accept", accept, 'str')

    return HttpRequest(
        method="DELETE",
        url=_url,
        params=_query_parameters,
        headers=_header_parameters,
        **kwargs
    )


def build_get_request(
    subscription_id,  # type: str
    resource_group_name,  # type: str
    workspace_name,  # type: str
    name,  # type: str
    **kwargs  # type: Any
):
    # type: (...) -> HttpRequest
    api_version = kwargs.pop('api_version', "2024-04-01-preview")  # type: str

    accept = "application/json"
    # Construct URL
    _url = kwargs.pop("template_url", "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/components/{name}")  # pylint: disable=line-too-long
    path_format_arguments = {
        "subscriptionId": _SERIALIZER.url("subscription_id", subscription_id, 'str', min_length=1),
        "resourceGroupName": _SERIALIZER.url("resource_group_name", resource_group_name, 'str', max_length=90, min_length=1),
        "workspaceName": _SERIALIZER.url("workspace_name", workspace_name, 'str', pattern=r'^[a-zA-Z0-9][a-zA-Z0-9_-]{2,32}$'),
        "name": _SERIALIZER.url("name", name, 'str'),
    }

    _url = _format_url_section(_url, **path_format_arguments)

    # Construct parameters
    _query_parameters = kwargs.pop("params", {})  # type: Dict[str, Any]
    _query_parameters['api-version'] = _SERIALIZER.query("api_version", api_version, 'str')

    # Construct headers
    _header_parameters = kwargs.pop("headers", {})  # type: Dict[str, Any]
    _header_parameters['Accept'] = _SERIALIZER.header("accept", accept, 'str')

    return HttpRequest(
        method="GET",
        url=_url,
        params=_query_parameters,
        headers=_header_parameters,
        **kwargs
    )


def build_create_or_update_request(
    subscription_id,  # type: str
    resource_group_name,  # type: str
    workspace_name,  # type: str
    name,  # type: str
    **kwargs  # type: Any
):
    # type: (...) -> HttpRequest
    api_version = kwargs.pop('api_version', "2024-04-01-preview")  # type: str
    content_type = kwargs.pop('content_type', None)  # type: Optional[str]

    accept = "application/json"
    # Construct URL
    _url = kwargs.pop("template_url", "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/components/{name}")  # pylint: disable=line-too-long
    path_format_arguments = {
        "subscriptionId": _SERIALIZER.url("subscription_id", subscription_id, 'str', min_length=1),
        "resourceGroupName": _SERIALIZER.url("resource_group_name", resource_group_name, 'str', max_length=90, min_length=1),
        "workspaceName": _SERIALIZER.url("workspace_name", workspace_name, 'str', pattern=r'^[a-zA-Z0-9][a-zA-Z0-9_-]{2,32}$'),
        "name": _SERIALIZER.url("name", name, 'str', pattern=r'^[a-zA-Z0-9][a-zA-Z0-9\-_]{0,254}$'),
    }

    _url = _format_url_section(_url, **path_format_arguments)

    # Construct parameters
    _query_parameters = kwargs.pop("params", {})  # type: Dict[str, Any]
    _query_parameters['api-version'] = _SERIALIZER.query("api_version", api_version, 'str')

    # Construct headers
    _header_parameters = kwargs.pop("headers", {})  # type: Dict[str, Any]
    if content_type is not None:
        _header_parameters['Content-Type'] = _SERIALIZER.header("content_type", content_type, 'str')
    _header_parameters['Accept'] = _SERIALIZER.header("accept", accept, 'str')

    return HttpRequest(
        method="PUT",
        url=_url,
        params=_query_parameters,
        headers=_header_parameters,
        **kwargs
    )

# fmt: on
class ComponentContainersOperations(object):
    """ComponentContainersOperations operations.

    You should not instantiate this class directly. Instead, you should create a Client instance that
    instantiates it for you and attaches it as an attribute.

    :ivar models: Alias to model classes used in this operation group.
    :type models: ~azure.mgmt.machinelearningservices.models
    :param client: Client for service requests.
    :param config: Configuration of service client.
    :param serializer: An object model serializer.
    :param deserializer: An object model deserializer.
    """

    models = _models

    def __init__(self, client, config, serializer, deserializer):
        self._client = client
        self._serialize = serializer
        self._deserialize = deserializer
        self._config = config

    @distributed_trace
    def list(
        self,
        resource_group_name,  # type: str
        workspace_name,  # type: str
        skip=None,  # type: Optional[str]
        list_view_type=None,  # type: Optional[Union[str, "_models.ListViewType"]]
        **kwargs  # type: Any
    ):
        # type: (...) -> Iterable["_models.ComponentContainerResourceArmPaginatedResult"]
        """List component containers.

        List component containers.

        :param resource_group_name: The name of the resource group. The name is case insensitive.
        :type resource_group_name: str
        :param workspace_name: Name of Azure Machine Learning workspace.
        :type workspace_name: str
        :param skip: Continuation token for pagination.
        :type skip: str
        :param list_view_type: View type for including/excluding (for example) archived entities.
        :type list_view_type: str or ~azure.mgmt.machinelearningservices.models.ListViewType
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: An iterator like instance of either ComponentContainerResourceArmPaginatedResult or
         the result of cls(response)
        :rtype:
         ~azure.core.paging.ItemPaged[~azure.mgmt.machinelearningservices.models.ComponentContainerResourceArmPaginatedResult]
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        api_version = kwargs.pop('api_version', "2024-04-01-preview")  # type: str

        cls = kwargs.pop('cls', None)  # type: ClsType["_models.ComponentContainerResourceArmPaginatedResult"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))
        def prepare_request(next_link=None):
            if not next_link:
                
                request = build_list_request(
                    subscription_id=self._config.subscription_id,
                    resource_group_name=resource_group_name,
                    workspace_name=workspace_name,
                    api_version=api_version,
                    skip=skip,
                    list_view_type=list_view_type,
                    template_url=self.list.metadata['url'],
                )
                request = _convert_request(request)
                request.url = self._client.format_url(request.url)

            else:
                
                request = build_list_request(
                    subscription_id=self._config.subscription_id,
                    resource_group_name=resource_group_name,
                    workspace_name=workspace_name,
                    api_version=api_version,
                    skip=skip,
                    list_view_type=list_view_type,
                    template_url=next_link,
                )
                request = _convert_request(request)
                request.url = self._client.format_url(request.url)
                request.method = "GET"
            return request

        def extract_data(pipeline_response):
            deserialized = self._deserialize("ComponentContainerResourceArmPaginatedResult", pipeline_response)
            list_of_elem = deserialized.value
            if cls:
                list_of_elem = cls(list_of_elem)
            return deserialized.next_link or None, iter(list_of_elem)

        def get_next(next_link=None):
            request = prepare_request(next_link)

            pipeline_response = self._client._pipeline.run(  # pylint: disable=protected-access
                request,
                stream=False,
                **kwargs
            )
            response = pipeline_response.http_response

            if response.status_code not in [200]:
                map_error(status_code=response.status_code, response=response, error_map=error_map)
                error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
                raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

            return pipeline_response


        return ItemPaged(
            get_next, extract_data
        )
    list.metadata = {'url': "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/components"}  # type: ignore

    @distributed_trace
    def delete(  # pylint: disable=inconsistent-return-statements
        self,
        resource_group_name,  # type: str
        workspace_name,  # type: str
        name,  # type: str
        **kwargs  # type: Any
    ):
        # type: (...) -> None
        """Delete container.

        Delete container.

        :param resource_group_name: The name of the resource group. The name is case insensitive.
        :type resource_group_name: str
        :param workspace_name: Name of Azure Machine Learning workspace.
        :type workspace_name: str
        :param name: Container name.
        :type name: str
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: None, or the result of cls(response)
        :rtype: None
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType[None]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))

        api_version = kwargs.pop('api_version', "2024-04-01-preview")  # type: str

        
        request = build_delete_request(
            subscription_id=self._config.subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            name=name,
            api_version=api_version,
            template_url=self.delete.metadata['url'],
        )
        request = _convert_request(request)
        request.url = self._client.format_url(request.url)

        pipeline_response = self._client._pipeline.run(  # pylint: disable=protected-access
            request,
            stream=False,
            **kwargs
        )
        response = pipeline_response.http_response

        if response.status_code not in [200, 204]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
            raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

        if cls:
            return cls(pipeline_response, None, {})

    delete.metadata = {'url': "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/components/{name}"}  # type: ignore


    @distributed_trace
    def get(
        self,
        resource_group_name,  # type: str
        workspace_name,  # type: str
        name,  # type: str
        **kwargs  # type: Any
    ):
        # type: (...) -> "_models.ComponentContainer"
        """Get container.

        Get container.

        :param resource_group_name: The name of the resource group. The name is case insensitive.
        :type resource_group_name: str
        :param workspace_name: Name of Azure Machine Learning workspace.
        :type workspace_name: str
        :param name: Container name.
        :type name: str
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: ComponentContainer, or the result of cls(response)
        :rtype: ~azure.mgmt.machinelearningservices.models.ComponentContainer
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType["_models.ComponentContainer"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))

        api_version = kwargs.pop('api_version', "2024-04-01-preview")  # type: str

        
        request = build_get_request(
            subscription_id=self._config.subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            name=name,
            api_version=api_version,
            template_url=self.get.metadata['url'],
        )
        request = _convert_request(request)
        request.url = self._client.format_url(request.url)

        pipeline_response = self._client._pipeline.run(  # pylint: disable=protected-access
            request,
            stream=False,
            **kwargs
        )
        response = pipeline_response.http_response

        if response.status_code not in [200]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
            raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

        deserialized = self._deserialize('ComponentContainer', pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    get.metadata = {'url': "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/components/{name}"}  # type: ignore


    @distributed_trace
    def create_or_update(
        self,
        resource_group_name,  # type: str
        workspace_name,  # type: str
        name,  # type: str
        body,  # type: "_models.ComponentContainer"
        **kwargs  # type: Any
    ):
        # type: (...) -> "_models.ComponentContainer"
        """Create or update container.

        Create or update container.

        :param resource_group_name: The name of the resource group. The name is case insensitive.
        :type resource_group_name: str
        :param workspace_name: Name of Azure Machine Learning workspace.
        :type workspace_name: str
        :param name: Container name.
        :type name: str
        :param body: Container entity to create or update.
        :type body: ~azure.mgmt.machinelearningservices.models.ComponentContainer
        :keyword callable cls: A custom type or function that will be passed the direct response
        :return: ComponentContainer, or the result of cls(response)
        :rtype: ~azure.mgmt.machinelearningservices.models.ComponentContainer
        :raises: ~azure.core.exceptions.HttpResponseError
        """
        cls = kwargs.pop('cls', None)  # type: ClsType["_models.ComponentContainer"]
        error_map = {
            401: ClientAuthenticationError, 404: ResourceNotFoundError, 409: ResourceExistsError
        }
        error_map.update(kwargs.pop('error_map', {}))

        api_version = kwargs.pop('api_version', "2024-04-01-preview")  # type: str
        content_type = kwargs.pop('content_type', "application/json")  # type: Optional[str]

        _json = self._serialize.body(body, 'ComponentContainer')

        request = build_create_or_update_request(
            subscription_id=self._config.subscription_id,
            resource_group_name=resource_group_name,
            workspace_name=workspace_name,
            name=name,
            api_version=api_version,
            content_type=content_type,
            json=_json,
            template_url=self.create_or_update.metadata['url'],
        )
        request = _convert_request(request)
        request.url = self._client.format_url(request.url)

        pipeline_response = self._client._pipeline.run(  # pylint: disable=protected-access
            request,
            stream=False,
            **kwargs
        )
        response = pipeline_response.http_response

        if response.status_code not in [200, 201]:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = self._deserialize.failsafe_deserialize(_models.ErrorResponse, pipeline_response)
            raise HttpResponseError(response=response, model=error, error_format=ARMErrorFormat)

        if response.status_code == 200:
            deserialized = self._deserialize('ComponentContainer', pipeline_response)

        if response.status_code == 201:
            deserialized = self._deserialize('ComponentContainer', pipeline_response)

        if cls:
            return cls(pipeline_response, deserialized, {})

        return deserialized

    create_or_update.metadata = {'url': "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.MachineLearningServices/workspaces/{workspaceName}/components/{name}"}  # type: ignore




# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

import json
import logging
import os
import re
import time
import typing
from collections import Counter
from typing import Any, Dict, List, Optional, Tuple, Union

from marshmallow import Schema

from azure.ai.ml._restclient.v2022_10_01.models import ComponentVersion, ComponentVersionProperties
from azure.ai.ml._schema import PathAwareSchema
from azure.ai.ml._schema.pipeline.pipeline_component import PipelineComponentSchema
from azure.ai.ml._utils._asset_utils import get_object_hash
from azure.ai.ml._utils.utils import hash_dict, is_data_binding_expression
from azure.ai.ml.constants._common import ARM_ID_PREFIX, ASSET_ARM_ID_REGEX_FORMAT, COMPONENT_TYPE
from azure.ai.ml.constants._component import ComponentSource, NodeType
from azure.ai.ml.constants._job.pipeline import ValidationErrorCode
from azure.ai.ml.entities._builders import BaseNode, Command
from azure.ai.ml.entities._builders.control_flow_node import ControlFlowNode, LoopNode
from azure.ai.ml.entities._component.component import Component
from azure.ai.ml.entities._inputs_outputs import GroupInput, Input
from azure.ai.ml.entities._job.automl.automl_job import AutoMLJob
from azure.ai.ml.entities._job.pipeline._attr_dict import has_attr_safe, try_get_non_arbitrary_attr
from azure.ai.ml.entities._job.pipeline._pipeline_expression import PipelineExpression
from azure.ai.ml.entities._validation import MutableValidationResult
from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, ValidationException

module_logger = logging.getLogger(__name__)


class PipelineComponent(Component):
    """Pipeline component, currently used to store components in an azure.ai.ml.dsl.pipeline.

    :param name: Name of the component.
    :type name: str
    :param version: Version of the component.
    :type version: str
    :param description: Description of the component.
    :type description: str
    :param tags: Tag dictionary. Tags can be added, removed, and updated.
    :type tags: dict
    :param display_name: Display name of the component.
    :type display_name: str
    :param inputs: Component inputs.
    :type inputs: dict
    :param outputs: Component outputs.
    :type outputs: dict
    :param jobs: Id to components dict inside the pipeline definition.
    :type jobs: Dict[str, ~azure.ai.ml.entities._builders.BaseNode]
    :param is_deterministic: Whether the pipeline component is deterministic.
    :type is_deterministic: bool
    :raises ~azure.ai.ml.exceptions.ValidationException: Raised if PipelineComponent cannot be successfully validated.
        Details will be provided in the error message.
    """

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        version: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[Dict] = None,
        display_name: Optional[str] = None,
        inputs: Optional[Dict] = None,
        outputs: Optional[Dict] = None,
        jobs: Optional[Dict[str, BaseNode]] = None,
        is_deterministic: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        kwargs[COMPONENT_TYPE] = NodeType.PIPELINE
        super().__init__(
            name=name,
            version=version,
            description=description,
            tags=tags,
            display_name=display_name,
            inputs=inputs,
            outputs=outputs,
            is_deterministic=is_deterministic,  # type: ignore[arg-type]
            **kwargs,
        )
        self._jobs = self._process_jobs(jobs) if jobs else {}
        # for telemetry
        self._job_types, self._job_sources = self._get_job_type_and_source()
        # Private support: create pipeline component from pipeline job
        self._source_job_id = kwargs.pop("source_job_id", None)
        # TODO: set anonymous hash for reuse

    def _process_jobs(self, jobs: Dict[str, BaseNode]) -> Dict[str, BaseNode]:
        """Process and validate jobs.

        :param jobs: A map of node name to node
        :type jobs: Dict[str, BaseNode]
        :return: The processed jobs
        :rtype: Dict[str, BaseNode]
        """
        # Remove swept Command
        node_names_to_skip = []
        for node_name, job_instance in jobs.items():
            if isinstance(job_instance, Command) and job_instance._swept is True:
                node_names_to_skip.append(node_name)

        for key in node_names_to_skip:
            del jobs[key]

        # Set path and validate node type.
        for _, job_instance in jobs.items():
            if isinstance(job_instance, BaseNode):
                job_instance._set_base_path(self.base_path)

            if not isinstance(job_instance, (BaseNode, AutoMLJob, ControlFlowNode)):
                msg = f"Not supported pipeline job type: {type(job_instance)}"
                raise ValidationException(
                    message=msg,
                    no_personal_data_message=msg,
                    target=ErrorTarget.PIPELINE,
                    error_category=ErrorCategory.USER_ERROR,
                )
        return jobs

    def _customized_validate(self) -> MutableValidationResult:
        """Validate pipeline component structure.

        :return: The validation result
        :rtype: MutableValidationResult
        """
        validation_result = super(PipelineComponent, self)._customized_validate()

        # Validate inputs
        for input_name, input_value in self.inputs.items():
            if input_value.type is None:
                validation_result.append_error(
                    yaml_path="inputs.{}".format(input_name),
                    message="Parameter type unknown, please add type annotation or specify input default value.",
                    error_code=ValidationErrorCode.PARAMETER_TYPE_UNKNOWN,
                )

        # Validate all nodes
        for node_name, node in self.jobs.items():
            if isinstance(node, BaseNode):
                # Node inputs will be validated.
                validation_result.merge_with(node._validate(), "jobs.{}".format(node_name))
                if isinstance(node.component, Component):
                    # Validate binding if not remote resource.
                    validation_result.merge_with(self._validate_binding_inputs(node))
            elif isinstance(node, AutoMLJob):
                pass
            elif isinstance(node, ControlFlowNode):
                # Validate control flow node.
                validation_result.merge_with(node._validate(), "jobs.{}".format(node_name))
            else:
                validation_result.append_error(
                    yaml_path="jobs.{}".format(node_name),
                    message=f"Not supported pipeline job type: {type(node)}",
                )

        return validation_result

    def _validate_compute_is_set(self, *, parent_node_name: Optional[str] = None) -> MutableValidationResult:
        """Validate compute in pipeline component.

        This function will only be called from pipeline_job._validate_compute_is_set
        when both of the pipeline_job.compute and pipeline_job.settings.default_compute is None.
        Rules:
        - For pipeline node: will call node._component._validate_compute_is_set to validate node compute in sub graph.
        - For general node:
            - If _skip_required_compute_missing_validation is True, validation will be skipped.
            - All the rest of cases without compute will add compute not set error to validation result.

        :keyword parent_node_name: The name of the parent node.
        :type parent_node_name: Optional[str]
        :return: The validation result
        :rtype: MutableValidationResult
        """

        # Note: do not put this into customized validate, as we would like call
        # this from pipeline_job._validate_compute_is_set
        validation_result = self._create_empty_validation_result()
        no_compute_nodes = []
        parent_node_name = parent_node_name if parent_node_name else ""
        for node_name, node in self.jobs.items():
            full_node_name = f"{parent_node_name}{node_name}.jobs."
            if node.type == NodeType.PIPELINE and isinstance(node._component, PipelineComponent):
                validation_result.merge_with(node._component._validate_compute_is_set(parent_node_name=full_node_name))
                continue
            if isinstance(node, BaseNode) and node._skip_required_compute_missing_validation:
                continue
            if has_attr_safe(node, "compute") and node.compute is None:
                no_compute_nodes.append(node_name)

        for node_name in no_compute_nodes:
            validation_result.append_error(
                yaml_path=f"jobs.{parent_node_name}{node_name}.compute",
                message="Compute not set",
            )
        return validation_result

    def _get_input_binding_dict(self, node: BaseNode) -> Tuple[dict, dict]:
        """Return the input binding dict for each node.

        :param node: The node
        :type node: BaseNode
        :return: A 2-tuple of (binding_dict, optional_binding_in_expression_dict)
        :rtype: Tuple[dict, dict]
        """
        # pylint: disable=too-many-nested-blocks
        binding_inputs = node._build_inputs()
        # Collect binding relation dict {'pipeline_input': ['node_input']}
        binding_dict: dict = {}
        optional_binding_in_expression_dict: dict = {}
        for component_input_name, component_binding_input in binding_inputs.items():
            if isinstance(component_binding_input, PipelineExpression):
                for pipeline_input_name in component_binding_input._inputs.keys():
                    if pipeline_input_name not in self.inputs:
                        continue
                    if pipeline_input_name not in binding_dict:
                        binding_dict[pipeline_input_name] = []
                    binding_dict[pipeline_input_name].append(component_input_name)
                    if pipeline_input_name not in optional_binding_in_expression_dict:
                        optional_binding_in_expression_dict[pipeline_input_name] = []
                    optional_binding_in_expression_dict[pipeline_input_name].append(pipeline_input_name)
            else:
                if isinstance(component_binding_input, Input):
                    component_binding_input = component_binding_input.path
                if is_data_binding_expression(component_binding_input, ["parent"]):
                    # data binding may have more than one PipelineInput now
                    for pipeline_input_name in PipelineExpression.parse_pipeline_inputs_from_data_binding(
                        component_binding_input
                    ):
                        if pipeline_input_name not in self.inputs:
                            continue
                        if pipeline_input_name not in binding_dict:
                            binding_dict[pipeline_input_name] = []
                        binding_dict[pipeline_input_name].append(component_input_name)
                        # for data binding expression "${{parent.inputs.pipeline_input}}", it should not be optional
                        if len(component_binding_input.replace("${{parent.inputs." + pipeline_input_name + "}}", "")):
                            if pipeline_input_name not in optional_binding_in_expression_dict:
                                optional_binding_in_expression_dict[pipeline_input_name] = []
                            optional_binding_in_expression_dict[pipeline_input_name].append(pipeline_input_name)
        return binding_dict, optional_binding_in_expression_dict

    def _validate_binding_inputs(self, node: BaseNode) -> MutableValidationResult:
        """Validate pipeline binding inputs and return all used pipeline input names.

        Mark input as optional if all binding is optional and optional not set. Raise error if pipeline input is
        optional but link to required inputs.

        :param node: The node to validate
        :type node: BaseNode
        :return: The validation result
        :rtype: MutableValidationResult
        """
        component_definition_inputs = {}
        # Add flattened group input into definition inputs.
        # e.g. Add {'group_name.item': PipelineInput} for {'group_name': GroupInput}
        for name, val in node.component.inputs.items():
            if isinstance(val, GroupInput):
                component_definition_inputs.update(val.flatten(group_parameter_name=name))
            component_definition_inputs[name] = val
        # Collect binding relation dict {'pipeline_input': ['node_input']}
        validation_result = self._create_empty_validation_result()
        binding_dict, optional_binding_in_expression_dict = self._get_input_binding_dict(node)

        # Validate links required and optional
        for pipeline_input_name, binding_inputs in binding_dict.items():
            pipeline_input = self.inputs[pipeline_input_name]
            required_bindings = []
            for name in binding_inputs:
                # not check optional/required for pipeline input used in pipeline expression
                if name in optional_binding_in_expression_dict.get(pipeline_input_name, []):
                    continue
                if name in component_definition_inputs and component_definition_inputs[name].optional is not True:
                    required_bindings.append(f"{node.name}.inputs.{name}")
            if pipeline_input.optional is None and not required_bindings:
                # Set input as optional if all binding is optional and optional not set.
                pipeline_input.optional = True
                pipeline_input._is_inferred_optional = True
            elif pipeline_input.optional is True and required_bindings:
                if pipeline_input._is_inferred_optional:
                    # Change optional=True to None if is inferred by us
                    pipeline_input.optional = None
                else:
                    # Raise exception if pipeline input is optional set by user but link to required inputs.
                    validation_result.append_error(
                        yaml_path="inputs.{}".format(pipeline_input._port_name),
                        message=f"Pipeline optional Input binding to required inputs: {required_bindings}",
                    )
        return validation_result

    def _get_job_type_and_source(self) -> Tuple[Dict[str, int], Dict[str, int]]:
        """Get job types and sources for telemetry.

        :return: A 2-tuple of
          * A map of job type to the number of occurrences
          * A map of job source to the number of occurrences
        :rtype: Tuple[Dict[str, int], Dict[str, int]]
        """
        job_types: list = []
        job_sources = []
        for job in self.jobs.values():
            job_types.append(job.type)
            if isinstance(job, BaseNode):
                job_sources.append(job._source)
            elif isinstance(job, AutoMLJob):
                # Consider all automl_job has builder type for now,
                # as it's not easy to distinguish their source(yaml/builder).
                job_sources.append(ComponentSource.BUILDER)
            else:
                # Fall back to CLASS
                job_sources.append(ComponentSource.CLASS)
        return dict(Counter(job_types)), dict(Counter(job_sources))

    @property
    def jobs(self) -> Dict[str, BaseNode]:
        """Return a dictionary from component variable name to component object.

        :return: Dictionary mapping component variable names to component objects.
        :rtype: Dict[str, ~azure.ai.ml.entities._builders.BaseNode]
        """
        return self._jobs

    def _get_anonymous_hash(self) -> str:
        """Get anonymous hash for pipeline component.

        :return: The anonymous hash of the pipeline component
        :rtype: str
        """
        # ideally we should always use rest object to generate hash as it's the same as
        # what we send to server-side, but changing the hash function will break reuse of
        # existing components except for command component (hash result is the same for
        # command component), so we just use rest object to generate hash for pipeline component,
        # which doesn't have reuse issue.
        component_interface_dict = self._to_rest_object().properties.component_spec
        # Hash local inputs in pipeline component jobs
        for job_name, job in self.jobs.items():
            if getattr(job, "inputs", None):
                for input_name, input_value in job.inputs.items():
                    try:
                        if (
                            getattr(input_value, "_data", None)
                            and isinstance(input_value._data, Input)
                            and input_value.path
                            and os.path.exists(input_value.path)
                        ):
                            start_time = time.time()
                            component_interface_dict["jobs"][job_name]["inputs"][input_name]["content_hash"] = (
                                get_object_hash(input_value.path)
                            )
                            module_logger.debug(
                                "Takes %s seconds to calculate the content hash of local input %s",
                                time.time() - start_time,
                                input_value.path,
                            )
                    except ValidationException:
                        pass
        hash_value: str = hash_dict(
            component_interface_dict,
            keys_to_omit=[
                # omit name since anonymous component will have same name
                "name",
                # omit _source since it doesn't impact component's uniqueness
                "_source",
                # omit id since it will be set after component is registered
                "id",
                # omit version since it will be set to this hash later
                "version",
            ],
        )
        return hash_value

    @classmethod
    def _load_from_rest_pipeline_job(cls, data: Dict) -> "PipelineComponent":
        # TODO: refine this?
        # Set type as None here to avoid schema validation failed
        definition_inputs = {p: {"type": None} for p in data.get("inputs", {}).keys()}
        definition_outputs = {p: {"type": None} for p in data.get("outputs", {}).keys()}
        return PipelineComponent(
            display_name=data.get("display_name"),
            description=data.get("description"),
            inputs=definition_inputs,
            outputs=definition_outputs,
            jobs=data.get("jobs"),
            _source=ComponentSource.REMOTE_WORKSPACE_JOB,
        )

    @classmethod
    def _resolve_sub_nodes(cls, rest_jobs: Dict) -> Dict:
        from azure.ai.ml.entities._job.pipeline._load_component import pipeline_node_factory

        sub_nodes = {}
        if rest_jobs is None:
            return sub_nodes
        for node_name, node in rest_jobs.items():
            # TODO: Remove this ad-hoc fix after unified arm id format in object
            component_id = node.get("componentId", "")
            if isinstance(component_id, str) and re.match(ASSET_ARM_ID_REGEX_FORMAT, component_id):
                node["componentId"] = component_id[len(ARM_ID_PREFIX) :]
            if not LoopNode._is_loop_node_dict(node):
                # skip resolve LoopNode first since it may reference other nodes
                # use node factory instead of BaseNode._from_rest_object here as AutoMLJob is not a BaseNode
                sub_nodes[node_name] = pipeline_node_factory.load_from_rest_object(obj=node)
        for node_name, node in rest_jobs.items():
            if LoopNode._is_loop_node_dict(node):
                # resolve LoopNode after all other nodes are resolved
                sub_nodes[node_name] = pipeline_node_factory.load_from_rest_object(obj=node, pipeline_jobs=sub_nodes)
        return sub_nodes

    @classmethod
    def _create_schema_for_validation(cls, context: Any) -> Union[PathAwareSchema, Schema]:
        return PipelineComponentSchema(context=context)

    @classmethod
    def _get_skip_fields_in_schema_validation(cls) -> typing.List[str]:
        # jobs validations are done in _customized_validate()
        return ["jobs"]

    @classmethod
    def _check_ignored_keys(cls, obj: object) -> List[str]:
        """Return ignored keys in obj as a pipeline component when its value be set.

        :param obj: The object to examine
        :type obj: object
        :return: List of keys to ignore
        :rtype: List[str]
        """
        examine_mapping = {
            "compute": lambda val: val is not None,
            "settings": lambda val: val is not None and any(v is not None for v in val._to_dict().values()),
        }
        # Avoid new attr added by use `try_get_non...` instead of `hasattr` or `getattr` directly.
        return [k for k, has_set in examine_mapping.items() if has_set(try_get_non_arbitrary_attr(obj, k))]

    def _get_telemetry_values(self, *args: Any, **kwargs: Any) -> Dict:
        telemetry_values: dict = super()._get_telemetry_values()
        telemetry_values.update(
            {
                "source": self._source,
                "node_count": len(self.jobs),
                "node_type": json.dumps(self._job_types),
                "node_source": json.dumps(self._job_sources),
            }
        )
        return telemetry_values

    @classmethod
    def _from_rest_object_to_init_params(cls, obj: ComponentVersion) -> Dict:
        # Pop jobs to avoid it goes with schema load
        jobs = obj.properties.component_spec.pop("jobs", None)
        init_params_dict: dict = super()._from_rest_object_to_init_params(obj)
        if jobs:
            try:
                init_params_dict["jobs"] = PipelineComponent._resolve_sub_nodes(jobs)
            except Exception as e:  # pylint: disable=W0718
                # Skip parse jobs if error exists.
                # TODO: https://msdata.visualstudio.com/Vienna/_workitems/edit/2052262
                module_logger.debug("Parse pipeline component jobs failed with: %s", e)
        return init_params_dict

    def _to_dict(self) -> Dict:
        return {**self._other_parameter, **super()._to_dict()}

    def _build_rest_component_jobs(self) -> Dict[str, dict]:
        """Build pipeline component jobs to rest.

        :return: A map of job name to rest objects
        :rtype: Dict[str, dict]
        """
        # Build the jobs to dict
        rest_component_jobs = {}
        for job_name, job in self.jobs.items():
            if isinstance(job, (BaseNode, ControlFlowNode)):
                rest_node_dict = job._to_rest_object()
            elif isinstance(job, AutoMLJob):
                rest_node_dict = json.loads(json.dumps(job._to_dict(inside_pipeline=True)))
            else:
                msg = f"Non supported job type in Pipeline jobs: {type(job)}"
                raise ValidationException(
                    message=msg,
                    no_personal_data_message=msg,
                    target=ErrorTarget.PIPELINE,
                    error_category=ErrorCategory.USER_ERROR,
                )
            rest_component_jobs[job_name] = rest_node_dict
        return rest_component_jobs

    def _to_rest_object(self) -> ComponentVersion:
        """Check ignored keys and return rest object.

        :return: The component version
        :rtype: ComponentVersion
        """
        ignored_keys = self._check_ignored_keys(self)
        if ignored_keys:
            module_logger.warning("%s ignored on pipeline component %r.", ignored_keys, self.name)
        component = self._to_dict()
        # add source type to component rest object
        component["_source"] = self._source
        component["jobs"] = self._build_rest_component_jobs()
        component["sourceJobId"] = self._source_job_id
        if self._intellectual_property:
            # hack while full pass through supported is worked on for IPP fields
            component.pop("intellectual_property")
            component["intellectualProperty"] = self._intellectual_property._to_rest_object().serialize()
        properties = ComponentVersionProperties(
            component_spec=component,
            description=self.description,
            is_anonymous=self._is_anonymous,
            properties=self.properties,
            tags=self.tags,
        )
        result = ComponentVersion(properties=properties)
        result.name = self.name
        return result

    def __str__(self) -> str:
        try:
            toYaml: str = self._to_yaml()
            return toYaml
        except BaseException:  # pylint: disable=W0718
            toStr: str = super(PipelineComponent, self).__str__()
            return toStr



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

import logging

from marshmallow import fields, post_load
from azure.ai.ml._utils.utils import camel_to_snake
from azure.ai.ml._schema.core.schema import PathAwareSchema
from azure.ai.ml._schema.core.fields import StringTransformedEnum, UnionField, NestedField

module_logger = logging.getLogger(__name__)


class PathBaseSchema(PathAwareSchema):
    input_path_type = StringTransformedEnum(
        allowed_values=[
            "path_id",
            "url",
            "path_version",
        ],
        casing_transform=camel_to_snake,
    )


class PackageInputPathIdSchema(PathBaseSchema):
    resource_id = fields.Str()

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._assets._artifacts._package.model_package import PackageInputPathId

        return PackageInputPathId(**data)


class PackageInputPathUrlSchema(PathBaseSchema):
    url = fields.Str()

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._assets._artifacts._package.model_package import PackageInputPathUrl

        return PackageInputPathUrl(**data)


class PackageInputPathSchema(PathBaseSchema):
    resource_name = fields.Str()
    resource_version = fields.Str()

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._assets._artifacts._package.model_package import PackageInputPathVersion

        return PackageInputPathVersion(**data)


class ModelPackageInputSchema(PathAwareSchema):
    type = StringTransformedEnum(allowed_values=["uri_file", "uri_folder"], casing_transform=camel_to_snake)
    mode = StringTransformedEnum(
        allowed_values=[
            "read_only_mount",
            "download",
        ],
        casing_transform=camel_to_snake,
    )
    path = UnionField(
        [
            NestedField(PackageInputPathIdSchema),
            NestedField(PackageInputPathUrlSchema),
            NestedField(PackageInputPathSchema),
        ]
    )
    mount_path = fields.Str()

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._assets._artifacts._package.model_package import ModelPackageInput

        return ModelPackageInput(**data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument,no-else-return

from marshmallow import EXCLUDE, fields
from marshmallow.decorators import post_load, pre_dump

from azure.ai.ml._schema import ExperimentalField
from azure.ai.ml._schema.core.fields import NestedField, StringTransformedEnum, UnionField
from azure.ai.ml._schema.core.schema_meta import PatchedSchemaMeta
from azure.ai.ml._utils.utils import _snake_to_camel, camel_to_snake
from azure.ai.ml.constants._workspace import FirewallSku, IsolationMode, OutboundRuleCategory
from azure.ai.ml.entities._workspace.networking import (
    FqdnDestination,
    ManagedNetwork,
    PrivateEndpointDestination,
    ServiceTagDestination,
)


class ManagedNetworkStatusSchema(metaclass=PatchedSchemaMeta):
    spark_ready = fields.Bool(dump_only=True)
    status = fields.Str(dump_only=True)


class FqdnOutboundRuleSchema(metaclass=PatchedSchemaMeta):
    name = fields.Str(required=True)
    parent_rule_names = fields.List(fields.Str(), dump_only=True)
    type = fields.Constant("fqdn")
    destination = fields.Str(required=True)
    category = StringTransformedEnum(
        allowed_values=[
            OutboundRuleCategory.REQUIRED,
            OutboundRuleCategory.RECOMMENDED,
            OutboundRuleCategory.USER_DEFINED,
        ],
        casing_transform=camel_to_snake,
        metadata={"description": "outbound rule category."},
        dump_only=True,
    )
    status = fields.Str(dump_only=True)

    @post_load
    def createdestobject(self, data, **kwargs):
        dest = data.get("destination")
        category = data.get("category", OutboundRuleCategory.USER_DEFINED)
        name = data.get("name")
        status = data.get("status", None)
        return FqdnDestination(
            name=name,
            destination=dest,
            category=_snake_to_camel(category),
            status=status,
        )


class ServiceTagDestinationSchema(metaclass=PatchedSchemaMeta):
    service_tag = fields.Str(required=True)
    protocol = fields.Str(required=True)
    port_ranges = fields.Str(required=True)
    address_prefixes = fields.List(fields.Str())


class ServiceTagOutboundRuleSchema(metaclass=PatchedSchemaMeta):
    name = fields.Str(required=True)
    parent_rule_names = fields.List(fields.Str(), dump_only=True)
    type = fields.Constant("service_tag")
    destination = NestedField(ServiceTagDestinationSchema, required=True)
    category = StringTransformedEnum(
        allowed_values=[
            OutboundRuleCategory.REQUIRED,
            OutboundRuleCategory.RECOMMENDED,
            OutboundRuleCategory.USER_DEFINED,
        ],
        casing_transform=camel_to_snake,
        metadata={"description": "outbound rule category."},
        dump_only=True,
    )
    status = fields.Str(dump_only=True)

    @pre_dump
    def predump(self, data, **kwargs):
        data.destination = self.service_tag_dest2dict(
            data.service_tag, data.protocol, data.port_ranges, data.address_prefixes
        )
        return data

    @post_load
    def createdestobject(self, data, **kwargs):
        dest = data.get("destination")
        category = data.get("category", OutboundRuleCategory.USER_DEFINED)
        name = data.get("name")
        status = data.get("status", None)
        return ServiceTagDestination(
            name=name,
            service_tag=dest["service_tag"],
            protocol=dest["protocol"],
            port_ranges=dest["port_ranges"],
            address_prefixes=dest.get("address_prefixes", None),
            category=_snake_to_camel(category),
            status=status,
        )

    def service_tag_dest2dict(self, service_tag, protocol, port_ranges, address_prefixes):
        service_tag_dest = {}
        service_tag_dest["service_tag"] = service_tag
        service_tag_dest["protocol"] = protocol
        service_tag_dest["port_ranges"] = port_ranges
        service_tag_dest["address_prefixes"] = address_prefixes
        return service_tag_dest


class PrivateEndpointDestinationSchema(metaclass=PatchedSchemaMeta):
    service_resource_id = fields.Str(required=True)
    subresource_target = fields.Str(required=True)
    spark_enabled = fields.Bool(required=True)


class PrivateEndpointOutboundRuleSchema(metaclass=PatchedSchemaMeta):
    name = fields.Str(required=True)
    parent_rule_names = fields.List(fields.Str(), dump_only=True)
    type = fields.Constant("private_endpoint")
    destination = NestedField(PrivateEndpointDestinationSchema, required=True)
    fqdns = fields.List(fields.Str())
    category = StringTransformedEnum(
        allowed_values=[
            OutboundRuleCategory.REQUIRED,
            OutboundRuleCategory.RECOMMENDED,
            OutboundRuleCategory.USER_DEFINED,
            OutboundRuleCategory.DEPENDENCY,
        ],
        casing_transform=camel_to_snake,
        metadata={"description": "outbound rule category."},
        dump_only=True,
    )
    status = fields.Str(dump_only=True)

    @pre_dump
    def predump(self, data, **kwargs):
        data.destination = self.pe_dest2dict(data.service_resource_id, data.subresource_target, data.spark_enabled)
        return data

    @post_load
    def createdestobject(self, data, **kwargs):
        dest = data.get("destination")
        category = data.get("category", OutboundRuleCategory.USER_DEFINED)
        name = data.get("name")
        status = data.get("status", None)
        fqdns = data.get("fqdns", None)
        return PrivateEndpointDestination(
            name=name,
            service_resource_id=dest["service_resource_id"],
            subresource_target=dest["subresource_target"],
            spark_enabled=dest["spark_enabled"],
            category=_snake_to_camel(category),
            status=status,
            fqdns=fqdns,
        )

    def pe_dest2dict(self, service_resource_id, subresource_target, spark_enabled):
        pedest = {}
        pedest["service_resource_id"] = service_resource_id
        pedest["subresource_target"] = subresource_target
        pedest["spark_enabled"] = spark_enabled
        return pedest


class ManagedNetworkSchema(metaclass=PatchedSchemaMeta):
    isolation_mode = StringTransformedEnum(
        allowed_values=[
            IsolationMode.DISABLED,
            IsolationMode.ALLOW_INTERNET_OUTBOUND,
            IsolationMode.ALLOW_ONLY_APPROVED_OUTBOUND,
        ],
        casing_transform=camel_to_snake,
        metadata={"description": "isolation mode for the workspace managed network."},
    )
    outbound_rules = fields.List(
        UnionField(
            [
                NestedField(PrivateEndpointOutboundRuleSchema, allow_none=False, unknown=EXCLUDE),
                NestedField(ServiceTagOutboundRuleSchema, allow_none=False, unknown=EXCLUDE),
                NestedField(
                    FqdnOutboundRuleSchema, allow_none=False, unknown=EXCLUDE
                ),  # this needs to be last since otherwise union field with match destination as a string
            ],
            allow_none=False,
            is_strict=True,
        ),
        allow_none=True,
    )
    firewall_sku = ExperimentalField(
        StringTransformedEnum(
            allowed_values=[
                FirewallSku.STANDARD,
                FirewallSku.BASIC,
            ],
            casing_transform=camel_to_snake,
            metadata={"description": "Firewall sku for FQDN rules in AllowOnlyApprovedOutbound mode"},
        )
    )
    network_id = fields.Str(required=False, dump_only=True)
    status = NestedField(ManagedNetworkStatusSchema, allow_none=False, unknown=EXCLUDE)

    @post_load
    def make(self, data, **kwargs):
        outbound_rules = data.get("outbound_rules", False)

        firewall_sku = data.get("firewall_sku", False)
        firewall_sku_value = _snake_to_camel(data["firewall_sku"]) if firewall_sku else FirewallSku.STANDARD

        if outbound_rules:
            return ManagedNetwork(
                isolation_mode=_snake_to_camel(data["isolation_mode"]),
                outbound_rules=outbound_rules,
                firewall_sku=firewall_sku_value,
            )
        else:
            return ManagedNetwork(
                isolation_mode=_snake_to_camel(data["isolation_mode"]),
                firewall_sku=firewall_sku_value,
            )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

"""A file for custom traceback for removing file path from the trace for compliance need."""

import os
import sys
import traceback
import types
from typing import Any, Generator, List, Optional, Type


class _CustomStackSummary(traceback.StackSummary):
    """Subclass of StackSummary."""

    def format(self) -> List[str]:
        """Format the stack ready for printing.

        :return: A list of strings ready for printing.
          * Each string in the resulting list corresponds to a single frame from the stack.
          * Each string ends in a newline; the strings may contain internal newlines as well, for those items with
            source text lines.
        :rtype: List[str]
        """
        result = []
        for frame in self:
            row = ['  File "{}", line {}, in {}\n'.format(os.path.basename(frame.filename), frame.lineno, frame.name)]
            if frame.line:
                row.append("    {}\n".format(frame.line.strip()))
            if frame.locals:
                for name, value in sorted(frame.locals.items()):
                    row.append("    {name} = {value}\n".format(name=name, value=value))
            result.append("".join(row))
        return result


# pylint: disable=too-many-instance-attributes
class _CustomTracebackException(traceback.TracebackException):
    # pylint: disable=super-init-not-called
    def __init__(
        self, exc_type, exc_value, exc_traceback, *, limit=None, lookup_lines=True, capture_locals=False, _seen=None
    ):
        if _seen is None:
            _seen = set()
        _seen.add(exc_value)
        # Gracefully handle (the way Python 2.4 and earlier did) the case of
        # being called with no type or value (None, None, None).
        if exc_value and exc_value.__cause__ is not None and exc_value.__cause__ not in _seen:
            cause = _CustomTracebackException(
                type(exc_value.__cause__),
                exc_value.__cause__,
                exc_value.__cause__.__traceback__,
                limit=limit,
                lookup_lines=False,
                capture_locals=capture_locals,
                _seen=_seen,
            )
        else:
            cause = None
        if exc_value and exc_value.__context__ is not None and exc_value.__context__ not in _seen:
            context = _CustomTracebackException(
                type(exc_value.__context__),
                exc_value.__context__,
                exc_value.__context__.__traceback__,
                limit=limit,
                lookup_lines=False,
                capture_locals=capture_locals,
                _seen=_seen,
            )
        else:
            context = None
        self.exc_traceback = exc_traceback
        self.__cause__ = cause
        self.__context__ = context
        self.__suppress_context__ = exc_value.__suppress_context__ if exc_value else False
        # TODO: locals.
        self.stack = _CustomStackSummary.extract(
            traceback.walk_tb(exc_traceback),
            limit=limit,
            lookup_lines=lookup_lines,
            capture_locals=capture_locals,
        )
        self.exc_type = exc_type
        # Capture now to permit freeing resources: only complication is in the
        # unofficial API _format_final_exc_line
        self._str = traceback._some_str(exc_value)
        if exc_type and issubclass(exc_type, SyntaxError):
            # Handle SyntaxError's specially
            self.filename = exc_value.filename
            self.lineno = str(exc_value.lineno)
            self.text = exc_value.text
            self.offset = exc_value.offset
            self.msg = exc_value.msg
        if lookup_lines:
            self._load_lines()

    def format_exception_only(self) -> Generator:
        """Format the exception part of the traceback.

        :return: An iterable of strings, each ending in a newline. Normally, the generator emits a single
           string; however, for SyntaxError exceptions, it emits several lines that (when printed) display
           detailed information about where the syntax error occurred. The message indicating which exception occurred
           is always the last string in the output.
        :rtype: Iterable[str]
        """
        if self.exc_type is None:
            yield traceback._format_final_exc_line(None, self._str)
            return

        stype = self.exc_type.__qualname__
        smod = self.exc_type.__module__
        if smod not in ("__main__", "builtins"):
            stype = smod + "." + stype

        if not issubclass(self.exc_type, SyntaxError):
            yield traceback._format_final_exc_line(stype, self._str)  # type: ignore[attr-defined]
            return

        # It was a syntax error; show exactly where the problem was found.
        filename = os.path.basename(self.filename) or "<string>"
        lineno = str(self.lineno) or "?"
        yield '  File "{}", line {}\n'.format(filename, lineno)

        badline = self.text
        offset = self.offset
        if badline is not None:
            yield "    {}\n".format(badline.strip())
            if offset is not None:
                caretspace: Any = badline.rstrip("\n")
                offset = min(len(caretspace), offset) - 1
                caretspace = caretspace[:offset].lstrip()
                # non-space whitespace (likes tabs) must be kept for alignment
                caretspace = ((c.isspace() and c or " ") for c in caretspace)
                yield "    {}^\n".format("".join(caretspace))
        msg = self.msg or "<no detail available>"
        yield "{}: {}\n".format(stype, msg)


def format_exc(limit: Optional[int] = None, chain: bool = True) -> str:
    """Like print_exc() but return a string.

    :param limit: None to include all frames or the number of frames to include.
    :type limit: Optional[int]
    :param chain: Whether to format __cause__ and __context__. Defaults to True
    :type chain: bool
    :return: The formatted exception string
    :rtype: str
    """
    return "".join(format_exception(*sys.exc_info(), limit=limit, chain=chain))  # type: ignore[misc]


def format_exception(  # pylint: disable=unused-argument
    etype: Type[BaseException],
    value: BaseException,
    tb: types.TracebackType,
    limit: Optional[int] = None,
    chain: bool = True,
) -> List[str]:
    """Format a stack trace and the exception information.

    The arguments have the same meaning as the corresponding arguments to print_exception().

    :param etype: The type of the exception
    :type etype: Type[BaseException]
    :param value: The exception
    :type value: BaseException
    :param tb: The exception traceback
    :type tb: types.TracebackType
    :param limit: None to include all frames or the number of frames to include.
    :type limit: Optional[int]
    :param chain: Whether to format __cause__ and __context__. Defaults to True
    :type chain: bool
    :return: A list of strings, each ending in a newline and some containing internal newlines.
      When these lines are concatenated and printed, exactly the same text is printed as does print_exception().
    :rtype: List[str]
    """
    # format_exception has ignored etype for some time, and code such as cgitb
    # passes in bogus values as a result. For compatibility with such code we
    # ignore it here (rather than in the new TracebackException API).
    return list(_CustomTracebackException(type(value), value, tb, limit=limit).format(chain=chain))



import argparse
import os
import pickle
from pathlib import Path

import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error, r2_score

parser = argparse.ArgumentParser("score")
parser.add_argument("--predictions", type=str, help="Path of predictions and actual data")
parser.add_argument("--model", type=str, help="Path to model")
parser.add_argument("--score_report", type=str, help="Path to score report")


args = parser.parse_args()

print("hello scoring world...")

lines = [
    f"Model path: {args.model}",
    f"Predictions path: {args.predictions}",
    f"Scoring output path: {args.score_report}",
]

for line in lines:
    print(line)

# Load the test data with predicted values

print("mounted_path files: ")
arr = os.listdir(args.predictions)

print(arr)
df_list = []
for filename in arr:
    print("reading file: %s ..." % filename)
    with open(os.path.join(args.predictions, filename), "r") as handle:
        # print (handle.read())
        input_df = pd.read_csv((Path(args.predictions) / filename))
        df_list.append(input_df)

test_data = df_list[0]

# Load the model from input port
model = pickle.load(open((Path(args.model) / "model.sav"), "rb"))

# Print the results of scoring the predictions against actual values in the test data
# The coefficients
print("Coefficients: \n", model.coef_)

actuals = test_data["actual_cost"]
predictions = test_data["predicted_cost"]

# The mean squared error
print("Mean squared error: %.2f" % mean_squared_error(actuals, predictions))
# The coefficient of determination: 1 is perfect prediction
print("Coefficient of determination: %.2f" % r2_score(actuals, predictions))
print("Model: ", model)

# Print score report to a text file
(Path(args.score_report) / "score.txt").write_text("Scored with the following model:\n{}".format(model))
with open((Path(args.score_report) / "score.txt"), "a") as f:
    f.write("\n Coefficients: \n %s \n" % str(model.coef_))
    f.write("Mean squared error: %.2f \n" % mean_squared_error(actuals, predictions))
    f.write("Coefficient of determination: %.2f \n" % r2_score(actuals, predictions))



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

import inspect
import logging
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, Optional, TypeVar, Union, overload

from typing_extensions import ParamSpec

from azure.ai.ml.entities import Data, Model, PipelineJob, PipelineJobSettings
from azure.ai.ml.entities._builders.pipeline import Pipeline
from azure.ai.ml.entities._inputs_outputs import Input
from azure.ai.ml.entities._job.pipeline._io import NodeOutput, PipelineInput, _GroupAttrDict
from azure.ai.ml.entities._job.pipeline._pipeline_expression import PipelineExpression
from azure.ai.ml.exceptions import UserErrorException

from azure.ai.ml.dsl._pipeline_component_builder import PipelineComponentBuilder, _is_inside_dsl_pipeline_func
from azure.ai.ml.dsl._pipeline_decorator import _validate_args
from azure.ai.ml.dsl._settings import _dsl_settings_stack
from azure.ai.ml.dsl._utils import _resolve_source_file

SUPPORTED_INPUT_TYPES = (
    PipelineInput,
    NodeOutput,
    Input,
    Model,
    Data,  # For the case use a Data object as an input, we will convert it to Input object
    Pipeline,  # For the case use a pipeline node as the input, we use its only one output as the real input.
    str,
    bool,
    int,
    float,
    PipelineExpression,
    _GroupAttrDict,
)
module_logger = logging.getLogger(__name__)

T = TypeVar("T")
P = ParamSpec("P")


# Overload the returns a decorator when func is None
@overload
def pipeline(
    func: None,
    *,
    name: Optional[str] = None,
    version: Optional[str] = None,
    display_name: Optional[str] = None,
    description: Optional[str] = None,
    experiment_name: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    **kwargs: Any,
) -> Callable[[Callable[P, T]], Callable[P, PipelineJob]]: ...


# Overload the returns a decorated function when func isn't None
@overload
def pipeline(
    func: Callable[P, T],
    *,
    name: Optional[str] = None,
    version: Optional[str] = None,
    display_name: Optional[str] = None,
    description: Optional[str] = None,
    experiment_name: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    **kwargs: Any,
) -> Callable[P, PipelineJob]: ...


def pipeline(
    func: Optional[Callable[P, T]] = None,
    *,
    name: Optional[str] = None,
    version: Optional[str] = None,
    display_name: Optional[str] = None,
    description: Optional[str] = None,
    experiment_name: Optional[str] = None,
    tags: Optional[Union[Dict[str, str], str]] = None,
    **kwargs: Any,
) -> Union[Callable[[Callable[P, T]], Callable[P, PipelineJob]], Callable[P, PipelineJob]]:
    """Build a pipeline which contains all component nodes defined in this function.

    :param func: The user pipeline function to be decorated.
    :type func: types.FunctionType
    :keyword name: The name of pipeline component, defaults to function name.
    :paramtype name: str
    :keyword version: The version of pipeline component, defaults to "1".
    :paramtype version: str
    :keyword display_name: The display name of pipeline component, defaults to function name.
    :paramtype display_name: str
    :keyword description: The description of the built pipeline.
    :paramtype description: str
    :keyword experiment_name: Name of the experiment the job will be created under, \
        if None is provided, experiment will be set to current directory.
    :paramtype experiment_name: str
    :keyword tags: The tags of pipeline component.
    :paramtype tags: dict[str, str]
    :return: Either
      * A decorator, if `func` is None
      * The decorated `func`

    :rtype: Union[
        Callable[[Callable], Callable[..., ~azure.ai.ml.entities.PipelineJob]],
        Callable[P, ~azure.ai.ml.entities.PipelineJob]

      ]

    .. admonition:: Example:

        .. literalinclude:: ../../../../samples/ml_samples_pipeline_job_configurations.py
            :start-after: [START configure_pipeline]
            :end-before: [END configure_pipeline]
            :language: python
            :dedent: 8
            :caption: Shows how to create a pipeline using this decorator.
    """

    # get_component force pipeline to return Pipeline instead of PipelineJob so we can set optional argument
    # need to remove get_component and rely on azure.ai.ml.dsl.pipeline
    get_component = kwargs.get("get_component", False)

    def pipeline_decorator(func: Callable[P, T]) -> Callable:
        if not isinstance(func, Callable):  # type: ignore
            raise UserErrorException(f"Dsl pipeline decorator accept only function type, got {type(func)}.")

        non_pipeline_inputs = kwargs.get("non_pipeline_inputs", []) or kwargs.get("non_pipeline_parameters", [])
        # compute variable names changed from default_compute_targe -> compute -> default_compute -> none
        # to support legacy usage, we support them with priority.
        compute = kwargs.get("compute", None)
        default_compute_target = kwargs.get("default_compute_target", None)
        default_compute_target = kwargs.get("default_compute", None) or default_compute_target
        continue_on_step_failure = kwargs.get("continue_on_step_failure", None)
        on_init = kwargs.get("on_init", None)
        on_finalize = kwargs.get("on_finalize", None)

        default_datastore = kwargs.get("default_datastore", None)
        force_rerun = kwargs.get("force_rerun", None)
        job_settings = {
            "default_datastore": default_datastore,
            "continue_on_step_failure": continue_on_step_failure,
            "force_rerun": force_rerun,
            "default_compute": default_compute_target,
            "on_init": on_init,
            "on_finalize": on_finalize,
        }
        func_entry_path = _resolve_source_file()
        if not func_entry_path:
            func_path = Path(inspect.getfile(func))
            # in notebook, func_path may be a fake path and will raise error when trying to resolve this fake path
            if func_path.exists():
                func_entry_path = func_path.resolve().absolute()

        job_settings = {k: v for k, v in job_settings.items() if v is not None}
        pipeline_builder = PipelineComponentBuilder(
            func=func,
            name=name,
            version=version,
            display_name=display_name,
            description=description,
            default_datastore=default_datastore,
            tags=tags,
            source_path=str(func_entry_path),
            non_pipeline_inputs=non_pipeline_inputs,
        )

        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Union[Pipeline, PipelineJob]:
            # Default args will be added here.
            # Node: push/pop stack here instead of put it inside build()
            # Because we only want to enable dsl settings on top level pipeline
            _dsl_settings_stack.push()  # use this stack to track on_init/on_finalize settings
            try:
                # Convert args to kwargs
                provided_positional_kwargs = _validate_args(func, args, kwargs, non_pipeline_inputs)

                # When pipeline supports variable params, update pipeline component to support the inputs in **kwargs.
                pipeline_parameters = {
                    k: v for k, v in provided_positional_kwargs.items() if k not in non_pipeline_inputs
                }
                pipeline_builder._update_inputs(pipeline_parameters)

                non_pipeline_params_dict = {
                    k: v for k, v in provided_positional_kwargs.items() if k in non_pipeline_inputs
                }

                # TODO: cache built pipeline component
                pipeline_component = pipeline_builder.build(
                    user_provided_kwargs=provided_positional_kwargs,
                    non_pipeline_inputs_dict=non_pipeline_params_dict,
                    non_pipeline_inputs=non_pipeline_inputs,
                )
            finally:
                # use `finally` to ensure pop operation from the stack
                dsl_settings = _dsl_settings_stack.pop()

            # update on_init/on_finalize settings if init/finalize job is set
            if dsl_settings.init_job_set:
                job_settings["on_init"] = dsl_settings.init_job_name(pipeline_component.jobs)
            if dsl_settings.finalize_job_set:
                job_settings["on_finalize"] = dsl_settings.finalize_job_name(pipeline_component.jobs)

            # TODO: pass compute & default_compute separately?
            common_init_args: Any = {
                "experiment_name": experiment_name,
                "component": pipeline_component,
                "inputs": pipeline_parameters,
                "tags": tags,
            }
            built_pipeline: Any = None
            if _is_inside_dsl_pipeline_func() or get_component:
                # on_init/on_finalize is not supported for pipeline component
                if job_settings.get("on_init") is not None or job_settings.get("on_finalize") is not None:
                    raise UserErrorException("On_init/on_finalize is not supported for pipeline component.")
                # Build pipeline node instead of pipeline job if inside dsl.
                built_pipeline = Pipeline(_from_component_func=True, **common_init_args)
                if job_settings:
                    module_logger.warning(
                        ("Job settings %s on pipeline function %r are ignored when using inside PipelineJob."),
                        job_settings,
                        func.__name__,
                    )
            else:
                built_pipeline = PipelineJob(
                    jobs=pipeline_component.jobs,
                    compute=compute,
                    settings=PipelineJobSettings(**job_settings),
                    **common_init_args,
                )

            return built_pipeline

        # Bug Item number: 2883169
        wrapper._is_dsl_func = True  # type: ignore
        wrapper._job_settings = job_settings  # type: ignore
        wrapper._pipeline_builder = pipeline_builder  # type: ignore
        return wrapper

    # enable use decorator without "()" if all arguments are default values
    if func is not None:
        return pipeline_decorator(func)
    return pipeline_decorator



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

from marshmallow import ValidationError, fields, post_load, pre_dump

from azure.ai.ml._schema._sweep.search_space.normal import NormalSchema, QNormalSchema
from azure.ai.ml._schema._sweep.search_space.randint import RandintSchema
from azure.ai.ml._schema._sweep.search_space.uniform import QUniformSchema, UniformSchema
from azure.ai.ml._schema.core.fields import (
    DumpableIntegerField,
    DumpableStringField,
    NestedField,
    StringTransformedEnum,
    UnionField,
)
from azure.ai.ml._schema.core.schema import PatchedSchemaMeta
from azure.ai.ml.constants._job.sweep import SearchSpace


class ChoiceSchema(metaclass=PatchedSchemaMeta):
    values = fields.List(
        UnionField(
            [
                DumpableIntegerField(strict=True),
                DumpableStringField(),
                fields.Float(),
                fields.Dict(
                    keys=fields.Str(),
                    values=UnionField(
                        [
                            NestedField("ChoiceSchema"),
                            NestedField(NormalSchema()),
                            NestedField(QNormalSchema()),
                            NestedField(RandintSchema()),
                            NestedField(UniformSchema()),
                            NestedField(QUniformSchema()),
                            DumpableIntegerField(strict=True),
                            fields.Float(),
                            fields.Str(),
                        ]
                    ),
                ),
            ]
        )
    )
    type = StringTransformedEnum(required=True, allowed_values=SearchSpace.CHOICE)

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.sweep import Choice

        return Choice(**data)

    @pre_dump
    def predump(self, data, **kwargs):
        from azure.ai.ml.sweep import Choice

        if not isinstance(data, Choice):
            raise ValidationError("Cannot dump non-Choice object into ChoiceSchema")
        return data



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument, name-too-long

from marshmallow import fields, post_load

from azure.ai.ml.constants._monitoring import MonitorFeatureType
from azure.ai.ml._schema.core.fields import StringTransformedEnum, NestedField
from azure.ai.ml._schema.core.schema import PatchedSchemaMeta


class MetricThresholdSchema(metaclass=PatchedSchemaMeta):
    threshold = fields.Number()


class NumericalDriftMetricsSchema(metaclass=PatchedSchemaMeta):
    jensen_shannon_distance = fields.Number()
    normalized_wasserstein_distance = fields.Number()
    population_stability_index = fields.Number()
    two_sample_kolmogorov_smirnov_test = fields.Number()

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import NumericalDriftMetrics

        return NumericalDriftMetrics(**data)


class CategoricalDriftMetricsSchema(metaclass=PatchedSchemaMeta):
    jensen_shannon_distance = fields.Number()
    population_stability_index = fields.Number()
    pearsons_chi_squared_test = fields.Number()

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import CategoricalDriftMetrics

        return CategoricalDriftMetrics(**data)


class DataDriftMetricThresholdSchema(MetricThresholdSchema):
    data_type = StringTransformedEnum(allowed_values=[MonitorFeatureType.NUMERICAL, MonitorFeatureType.CATEGORICAL])

    numerical = NestedField(NumericalDriftMetricsSchema)
    categorical = NestedField(CategoricalDriftMetricsSchema)

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import DataDriftMetricThreshold

        return DataDriftMetricThreshold(**data)


class DataQualityMetricsNumericalSchema(metaclass=PatchedSchemaMeta):
    null_value_rate = fields.Number()
    data_type_error_rate = fields.Number()
    out_of_bounds_rate = fields.Number()

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import DataQualityMetricsNumerical

        return DataQualityMetricsNumerical(**data)


class DataQualityMetricsCategoricalSchema(metaclass=PatchedSchemaMeta):
    null_value_rate = fields.Number()
    data_type_error_rate = fields.Number()
    out_of_bounds_rate = fields.Number()

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import DataQualityMetricsCategorical

        return DataQualityMetricsCategorical(**data)


class DataQualityMetricThresholdSchema(MetricThresholdSchema):
    data_type = StringTransformedEnum(allowed_values=[MonitorFeatureType.NUMERICAL, MonitorFeatureType.CATEGORICAL])
    numerical = NestedField(DataQualityMetricsNumericalSchema)
    categorical = NestedField(DataQualityMetricsCategoricalSchema)

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import DataQualityMetricThreshold

        return DataQualityMetricThreshold(**data)


class PredictionDriftMetricThresholdSchema(MetricThresholdSchema):
    data_type = StringTransformedEnum(allowed_values=[MonitorFeatureType.NUMERICAL, MonitorFeatureType.CATEGORICAL])
    numerical = NestedField(NumericalDriftMetricsSchema)
    categorical = NestedField(CategoricalDriftMetricsSchema)

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import PredictionDriftMetricThreshold

        return PredictionDriftMetricThreshold(**data)


# pylint: disable-next=name-too-long
class FeatureAttributionDriftMetricThresholdSchema(MetricThresholdSchema):
    normalized_discounted_cumulative_gain = fields.Number()

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import FeatureAttributionDriftMetricThreshold

        return FeatureAttributionDriftMetricThreshold(**data)


class ModelPerformanceClassificationThresholdsSchema(metaclass=PatchedSchemaMeta):
    accuracy = fields.Number()
    precision = fields.Number()
    recall = fields.Number()

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import ModelPerformanceClassificationThresholds

        return ModelPerformanceClassificationThresholds(**data)


class ModelPerformanceRegressionThresholdsSchema(metaclass=PatchedSchemaMeta):
    mae = fields.Number()
    mse = fields.Number()
    rmse = fields.Number()

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import ModelPerformanceRegressionThresholds

        return ModelPerformanceRegressionThresholds(**data)


class ModelPerformanceMetricThresholdSchema(MetricThresholdSchema):
    classification = NestedField(ModelPerformanceClassificationThresholdsSchema)
    regression = NestedField(ModelPerformanceRegressionThresholdsSchema)

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import ModelPerformanceMetricThreshold

        return ModelPerformanceMetricThreshold(**data)


class CustomMonitoringMetricThresholdSchema(MetricThresholdSchema):
    metric_name = fields.Str()

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import CustomMonitoringMetricThreshold

        return CustomMonitoringMetricThreshold(**data)


class GenerationSafetyQualityMetricThresholdSchema(metaclass=PatchedSchemaMeta):  # pylint: disable=name-too-long
    groundedness = fields.Dict(
        keys=StringTransformedEnum(
            allowed_values=["aggregated_groundedness_pass_rate", "acceptable_groundedness_score_per_instance"]
        ),
        values=fields.Number(),
    )
    relevance = fields.Dict(
        keys=StringTransformedEnum(
            allowed_values=["aggregated_relevance_pass_rate", "acceptable_relevance_score_per_instance"]
        ),
        values=fields.Number(),
    )
    coherence = fields.Dict(
        keys=StringTransformedEnum(
            allowed_values=["aggregated_coherence_pass_rate", "acceptable_coherence_score_per_instance"]
        ),
        values=fields.Number(),
    )
    fluency = fields.Dict(
        keys=StringTransformedEnum(
            allowed_values=["aggregated_fluency_pass_rate", "acceptable_fluency_score_per_instance"]
        ),
        values=fields.Number(),
    )
    similarity = fields.Dict(
        keys=StringTransformedEnum(
            allowed_values=["aggregated_similarity_pass_rate", "acceptable_similarity_score_per_instance"]
        ),
        values=fields.Number(),
    )

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import GenerationSafetyQualityMonitoringMetricThreshold

        return GenerationSafetyQualityMonitoringMetricThreshold(**data)


class GenerationTokenStatisticsMonitorMetricThresholdSchema(
    metaclass=PatchedSchemaMeta
):  # pylint: disable=name-too-long
    totaltoken = fields.Dict(
        keys=StringTransformedEnum(allowed_values=["total_token_count", "total_token_count_per_group"]),
        values=fields.Number(),
    )

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.thresholds import GenerationTokenStatisticsMonitorMetricThreshold

        return GenerationTokenStatisticsMonitorMetricThreshold(**data)



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

try:
    from ._models_py3 import Artifact
    from ._models_py3 import Asset
    from ._models_py3 import AssetDto
    from ._models_py3 import AssetPaginatedResult
    from ._models_py3 import BatchGetResolvedUrisDto
    from ._models_py3 import BatchModelPathResponseDto
    from ._models_py3 import BlobReference
    from ._models_py3 import BlobReferenceForConsumptionDto
    from ._models_py3 import ContainerResourceRequirements
    from ._models_py3 import CreateUnregisteredInputModelDto
    from ._models_py3 import CreateUnregisteredOutputModelDto
    from ._models_py3 import CreatedBy
    from ._models_py3 import CreationContext
    from ._models_py3 import DataItem
    from ._models_py3 import DataReferenceCredentialDto
    from ._models_py3 import DataReferences
    from ._models_py3 import DataReferencesForConsumptionDto
    from ._models_py3 import DatasetReference
    from ._models_py3 import DependencyMapDto
    from ._models_py3 import DependencyMapItemDto
    from ._models_py3 import DependentAsset
    from ._models_py3 import DependentEntitiesDto
    from ._models_py3 import ErrorResponse
    from ._models_py3 import ExtensiveModel
    from ._models_py3 import FeedIndexEntityDto
    from ._models_py3 import FeedIndexEntityRequestDto
    from ._models_py3 import ImageReference
    from ._models_py3 import ImageReferenceForConsumptionDto
    from ._models_py3 import IndexAnnotations
    from ._models_py3 import IndexEntity
    from ._models_py3 import IndexProperties
    from ._models_py3 import InnerErrorDetails
    from ._models_py3 import IntellectualPropertyPublisherInformation
    from ._models_py3 import ListModelsRequest
    from ._models_py3 import Model
    from ._models_py3 import ModelBatchDto
    from ._models_py3 import ModelBatchResponseDto
    from ._models_py3 import ModelContainerRequest
    from ._models_py3 import ModelDeploymentSettings
    from ._models_py3 import ModelListModelsRequestPagedResponse
    from ._models_py3 import ModelPagedResponse
    from ._models_py3 import ModelPathResponseDto
    from ._models_py3 import ModelSchema
    from ._models_py3 import ModelSettingsIdentifiers
    from ._models_py3 import Operation
    from ._models_py3 import ProviderFeedEntityRequestDto
    from ._models_py3 import Relationship
    from ._models_py3 import ServiceResponseBase
    from ._models_py3 import User
except (SyntaxError, ImportError):
    from ._models import Artifact  # type: ignore
    from ._models import Asset  # type: ignore
    from ._models import AssetDto  # type: ignore
    from ._models import AssetPaginatedResult  # type: ignore
    from ._models import BatchGetResolvedUrisDto  # type: ignore
    from ._models import BatchModelPathResponseDto  # type: ignore
    from ._models import BlobReference  # type: ignore
    from ._models import BlobReferenceForConsumptionDto  # type: ignore
    from ._models import ContainerResourceRequirements  # type: ignore
    from ._models import CreateUnregisteredInputModelDto  # type: ignore
    from ._models import CreateUnregisteredOutputModelDto  # type: ignore
    from ._models import CreatedBy  # type: ignore
    from ._models import CreationContext  # type: ignore
    from ._models import DataItem  # type: ignore
    from ._models import DataReferenceCredentialDto  # type: ignore
    from ._models import DataReferences  # type: ignore
    from ._models import DataReferencesForConsumptionDto  # type: ignore
    from ._models import DatasetReference  # type: ignore
    from ._models import DependencyMapDto  # type: ignore
    from ._models import DependencyMapItemDto  # type: ignore
    from ._models import DependentAsset  # type: ignore
    from ._models import DependentEntitiesDto  # type: ignore
    from ._models import ErrorResponse  # type: ignore
    from ._models import ExtensiveModel  # type: ignore
    from ._models import FeedIndexEntityDto  # type: ignore
    from ._models import FeedIndexEntityRequestDto  # type: ignore
    from ._models import ImageReference  # type: ignore
    from ._models import ImageReferenceForConsumptionDto  # type: ignore
    from ._models import IndexAnnotations  # type: ignore
    from ._models import IndexEntity  # type: ignore
    from ._models import IndexProperties  # type: ignore
    from ._models import InnerErrorDetails  # type: ignore
    from ._models import IntellectualPropertyPublisherInformation  # type: ignore
    from ._models import ListModelsRequest  # type: ignore
    from ._models import Model  # type: ignore
    from ._models import ModelBatchDto  # type: ignore
    from ._models import ModelBatchResponseDto  # type: ignore
    from ._models import ModelContainerRequest  # type: ignore
    from ._models import ModelDeploymentSettings  # type: ignore
    from ._models import ModelListModelsRequestPagedResponse  # type: ignore
    from ._models import ModelPagedResponse  # type: ignore
    from ._models import ModelPathResponseDto  # type: ignore
    from ._models import ModelSchema  # type: ignore
    from ._models import ModelSettingsIdentifiers  # type: ignore
    from ._models import Operation  # type: ignore
    from ._models import ProviderFeedEntityRequestDto  # type: ignore
    from ._models import Relationship  # type: ignore
    from ._models import ServiceResponseBase  # type: ignore
    from ._models import User  # type: ignore

from ._azure_machine_learning_workspaces_enums import (
    ComputeEnvironmentType,
    DeploymentType,
    EntityKind,
    ListViewType,
    ModelFormatEnum,
    ModelSchemaDataType,
    OrderString,
    WebServiceState,
)

__all__ = [
    'Artifact',
    'Asset',
    'AssetDto',
    'AssetPaginatedResult',
    'BatchGetResolvedUrisDto',
    'BatchModelPathResponseDto',
    'BlobReference',
    'BlobReferenceForConsumptionDto',
    'ContainerResourceRequirements',
    'CreateUnregisteredInputModelDto',
    'CreateUnregisteredOutputModelDto',
    'CreatedBy',
    'CreationContext',
    'DataItem',
    'DataReferenceCredentialDto',
    'DataReferences',
    'DataReferencesForConsumptionDto',
    'DatasetReference',
    'DependencyMapDto',
    'DependencyMapItemDto',
    'DependentAsset',
    'DependentEntitiesDto',
    'ErrorResponse',
    'ExtensiveModel',
    'FeedIndexEntityDto',
    'FeedIndexEntityRequestDto',
    'ImageReference',
    'ImageReferenceForConsumptionDto',
    'IndexAnnotations',
    'IndexEntity',
    'IndexProperties',
    'InnerErrorDetails',
    'IntellectualPropertyPublisherInformation',
    'ListModelsRequest',
    'Model',
    'ModelBatchDto',
    'ModelBatchResponseDto',
    'ModelContainerRequest',
    'ModelDeploymentSettings',
    'ModelListModelsRequestPagedResponse',
    'ModelPagedResponse',
    'ModelPathResponseDto',
    'ModelSchema',
    'ModelSettingsIdentifiers',
    'Operation',
    'ProviderFeedEntityRequestDto',
    'Relationship',
    'ServiceResponseBase',
    'User',
    'ComputeEnvironmentType',
    'DeploymentType',
    'EntityKind',
    'ListViewType',
    'ModelFormatEnum',
    'ModelSchemaDataType',
    'OrderString',
    'WebServiceState',
]



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from . import debug_mode
from .chained_identity import START_MSG, STOP_MSG, ChainedIdentity

__all__ = ["debug_mode", "ChainedIdentity", "START_MSG", "STOP_MSG"]



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
# pylint: disable=protected-access
import logging
from abc import ABC
from datetime import datetime
from typing import List, Optional, Union

from azure.ai.ml._restclient.v2023_04_01_preview.models import CronTrigger as RestCronTrigger
from azure.ai.ml._restclient.v2023_04_01_preview.models import RecurrenceSchedule as RestRecurrencePattern
from azure.ai.ml._restclient.v2023_04_01_preview.models import RecurrenceTrigger as RestRecurrenceTrigger
from azure.ai.ml._restclient.v2023_04_01_preview.models import TriggerBase as RestTriggerBase
from azure.ai.ml._restclient.v2023_04_01_preview.models import TriggerType as RestTriggerType
from azure.ai.ml._utils.utils import camel_to_snake, snake_to_camel
from azure.ai.ml.constants import TimeZone
from azure.ai.ml.entities._mixins import RestTranslatableMixin

module_logger = logging.getLogger(__name__)


class TriggerBase(RestTranslatableMixin, ABC):
    """Base class of Trigger.

    This class should not be instantiated directly. Instead, use one of its subclasses.

    :keyword type: The type of trigger.
    :paramtype type: str
    :keyword start_time: Specifies the start time of the schedule in ISO 8601 format.
    :paramtype start_time: Optional[Union[str, datetime]]
    :keyword end_time: Specifies the end time of the schedule in ISO 8601 format.
        Note that end_time is not supported for compute schedules.
    :paramtype end_time: Optional[Union[str, datetime]]
    :keyword time_zone: The time zone where the schedule will run. Defaults to UTC(+00:00).
        Note that this applies to the start_time and end_time.
    :paramtype time_zone: ~azure.ai.ml.constants.TimeZone
    """

    def __init__(
        self,
        *,
        type: str,  # pylint: disable=redefined-builtin
        start_time: Optional[Union[str, datetime]] = None,
        end_time: Optional[Union[str, datetime]] = None,
        time_zone: Union[str, TimeZone] = TimeZone.UTC,
    ) -> None:
        super().__init__()
        self.type = type
        self.start_time = start_time
        self.end_time = end_time
        self.time_zone = time_zone

    @classmethod
    def _from_rest_object(cls, obj: RestTriggerBase) -> Optional[Union["CronTrigger", "RecurrenceTrigger"]]:
        if obj.trigger_type == RestTriggerType.RECURRENCE:
            return RecurrenceTrigger._from_rest_object(obj)
        if obj.trigger_type == RestTriggerType.CRON:
            return CronTrigger._from_rest_object(obj)

        return None


class RecurrencePattern(RestTranslatableMixin):
    """Recurrence pattern for a job schedule.

    :keyword hours: The number of hours for the recurrence schedule pattern.
    :paramtype hours: Union[int, List[int]]
    :keyword minutes: The number of minutes for the recurrence schedule pattern.
    :paramtype minutes: Union[int, List[int]]
    :keyword week_days: A list of days of the week for the recurrence schedule pattern.
        Acceptable values include: "monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday"
    :type week_days: Optional[Union[str, List[str]]]
    :keyword month_days: A list of days of the month for the recurrence schedule pattern.
    :paramtype month_days: Optional[Union[int, List[int]]]

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_misc.py
            :start-after: [START job_schedule_configuration]
            :end-before: [END job_schedule_configuration]
            :language: python
            :dedent: 8
            :caption: Configuring a JobSchedule to use a RecurrencePattern.
    """

    def __init__(
        self,
        *,
        hours: Union[int, List[int]],
        minutes: Union[int, List[int]],
        week_days: Optional[Union[str, List[str]]] = None,
        month_days: Optional[Union[int, List[int]]] = None,
    ) -> None:
        self.hours = hours
        self.minutes = minutes
        self.week_days = week_days
        self.month_days = month_days

    def _to_rest_object(self) -> RestRecurrencePattern:
        return RestRecurrencePattern(
            hours=[self.hours] if not isinstance(self.hours, list) else self.hours,
            minutes=[self.minutes] if not isinstance(self.minutes, list) else self.minutes,
            week_days=[self.week_days] if self.week_days and not isinstance(self.week_days, list) else self.week_days,
            month_days=(
                [self.month_days] if self.month_days and not isinstance(self.month_days, list) else self.month_days
            ),
        )

    def _to_rest_compute_pattern_object(self) -> RestRecurrencePattern:
        # This function is added because we can't make compute trigger to use same class
        # with schedule from service side.
        if self.month_days:
            module_logger.warning("'month_days' is ignored for not supported on compute recurrence schedule.")
        return RestRecurrencePattern(
            hours=[self.hours] if not isinstance(self.hours, list) else self.hours,
            minutes=[self.minutes] if not isinstance(self.minutes, list) else self.minutes,
            week_days=[self.week_days] if self.week_days and not isinstance(self.week_days, list) else self.week_days,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestRecurrencePattern) -> "RecurrencePattern":
        return cls(
            hours=obj.hours,
            minutes=obj.minutes,
            week_days=obj.week_days,
            month_days=obj.month_days if hasattr(obj, "month_days") else None,
        )


class CronTrigger(TriggerBase):
    """Cron Trigger for a job schedule.

    :keyword expression: The cron expression of schedule, following NCronTab format.
    :paramtype expression: str
    :keyword start_time: The start time for the trigger. If using a datetime object, leave the tzinfo as None and use
        the ``time_zone`` parameter to specify a time zone if needed. If using a string, use the format
        YYYY-MM-DDThh:mm:ss. Defaults to running the first workload instantly and continuing future workloads
        based on the schedule. If the start time is in the past, the first workload is run at the next calculated run
        time.
    :paramtype start_time: Optional[Union[str, datetime]]
    :keyword end_time: The start time for the trigger. If using a datetime object, leave the tzinfo as None and use
        the ``time_zone`` parameter to specify a time zone if needed. If using a string, use the format
        YYYY-MM-DDThh:mm:ss. Note that end_time is not supported for compute schedules.
    :paramtype end_time: Optional[Union[str, datetime]]
    :keyword time_zone: The time zone where the schedule will run. Defaults to UTC(+00:00).
        Note that this applies to the start_time and end_time.
    :paramtype time_zone: Union[str, ~azure.ai.ml.constants.TimeZone]
    :raises Exception: Raised if end_time is in the past.

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_misc.py
            :start-after: [START cron_trigger_configuration]
            :end-before: [END cron_trigger_configuration]
            :language: python
            :dedent: 8
            :caption: Configuring a CronTrigger.
    """

    def __init__(
        self,
        *,
        expression: str,
        start_time: Optional[Union[str, datetime]] = None,
        end_time: Optional[Union[str, datetime]] = None,
        time_zone: Union[str, TimeZone] = TimeZone.UTC,
    ) -> None:
        super().__init__(
            type=RestTriggerType.CRON,
            start_time=start_time,
            end_time=end_time,
            time_zone=time_zone,
        )
        self.expression = expression

    def _to_rest_object(self) -> RestCronTrigger:  # v2022_12_01.models.CronTrigger
        return RestCronTrigger(
            trigger_type=self.type,
            expression=self.expression,
            start_time=self.start_time,
            end_time=self.end_time,
            time_zone=self.time_zone,
        )

    def _to_rest_compute_cron_object(self) -> RestCronTrigger:  # v2022_12_01_preview.models.CronTrigger
        # This function is added because we can't make compute trigger to use same class
        # with schedule from service side.
        if self.end_time:
            module_logger.warning("'end_time' is ignored for not supported on compute schedule.")
        return RestCronTrigger(
            expression=self.expression,
            start_time=self.start_time,
            time_zone=self.time_zone,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestCronTrigger) -> "CronTrigger":
        return cls(
            expression=obj.expression,
            start_time=obj.start_time,
            end_time=obj.end_time,
            time_zone=obj.time_zone,
        )


class RecurrenceTrigger(TriggerBase):
    """Recurrence trigger for a job schedule.

    :keyword start_time: Specifies the start time of the schedule in ISO 8601 format.
    :paramtype start_time: Optional[Union[str, datetime]]
    :keyword end_time: Specifies the end time of the schedule in ISO 8601 format.
        Note that end_time is not supported for compute schedules.
    :paramtype end_time: Optional[Union[str, datetime]]
    :keyword time_zone: The time zone where the schedule will run. Defaults to UTC(+00:00).
        Note that this applies to the start_time and end_time.
    :paramtype time_zone: Union[str, ~azure.ai.ml.constants.TimeZone]
    :keyword frequency: Specifies the frequency that the schedule should be triggered with.
     Possible values include: "minute", "hour", "day", "week", "month".
    :type frequency: str
    :keyword interval: Specifies the interval in conjunction with the frequency that the schedule should be triggered
        with.
    :paramtype interval: int
    :keyword schedule: Specifies the recurrence pattern.
    :paramtype schedule: Optional[~azure.ai.ml.entities.RecurrencePattern]

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_misc.py
            :start-after: [START job_schedule_configuration]
            :end-before: [END job_schedule_configuration]
            :language: python
            :dedent: 8
            :caption: Configuring a JobSchedule to trigger recurrence every 4 weeks.
    """

    def __init__(
        self,
        *,
        frequency: str,
        interval: int,
        schedule: Optional[RecurrencePattern] = None,
        start_time: Optional[Union[str, datetime]] = None,
        end_time: Optional[Union[str, datetime]] = None,
        time_zone: Union[str, TimeZone] = TimeZone.UTC,
    ) -> None:
        super().__init__(
            type=RestTriggerType.RECURRENCE,
            start_time=start_time,
            end_time=end_time,
            time_zone=time_zone,
        )
        # Create empty pattern as schedule is required in rest model
        self.schedule = schedule if schedule else RecurrencePattern(hours=[], minutes=[])
        self.frequency = frequency
        self.interval = interval

    def _to_rest_object(self) -> RestRecurrenceTrigger:  # v2022_12_01.models.RecurrenceTrigger
        return RestRecurrenceTrigger(
            frequency=snake_to_camel(self.frequency),
            interval=self.interval,
            schedule=self.schedule._to_rest_object(),
            start_time=self.start_time,
            end_time=self.end_time,
            time_zone=self.time_zone,
        )

    def _to_rest_compute_recurrence_object(self) -> RestRecurrenceTrigger:
        # v2022_12_01_preview.models.RecurrenceTrigger
        # This function is added because we can't make compute trigger to use same class
        # with schedule from service side.
        if self.end_time:
            module_logger.warning("'end_time' is ignored for not supported on compute schedule.")
        return RestRecurrenceTrigger(
            frequency=snake_to_camel(self.frequency),
            interval=self.interval,
            schedule=self.schedule._to_rest_compute_pattern_object(),
            start_time=self.start_time,
            time_zone=self.time_zone,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestRecurrenceTrigger) -> "RecurrenceTrigger":
        return cls(
            frequency=camel_to_snake(obj.frequency),
            interval=obj.interval,
            schedule=RecurrencePattern._from_rest_object(obj.schedule) if obj.schedule else None,
            start_time=obj.start_time,
            end_time=obj.end_time,
            time_zone=obj.time_zone,
        )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

from marshmallow import ValidationError, fields, post_load, pre_dump

from azure.ai.ml._schema import StringTransformedEnum
from azure.ai.ml._schema.core.schema_meta import PatchedSchemaMeta
from azure.ai.ml._utils._experimental import experimental
from azure.ai.ml.constants._registry import AcrAccountSku


@experimental
class SystemCreatedAcrAccountSchema(metaclass=PatchedSchemaMeta):
    arm_resource_id = fields.Str(dump_only=True)
    acr_account_sku = StringTransformedEnum(
        allowed_values=[sku.value for sku in AcrAccountSku], casing_transform=lambda x: x.lower()
    )

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities import SystemCreatedAcrAccount

        data.pop("type", None)
        return SystemCreatedAcrAccount(**data)

    @pre_dump
    def predump(self, data, **kwargs):
        from azure.ai.ml.entities import SystemCreatedAcrAccount

        if not isinstance(data, SystemCreatedAcrAccount):
            raise ValidationError("Cannot dump non-SystemCreatedAcrAccount object into SystemCreatedAcrAccountSchema")
        return data



import pytest

from azure.ai.ml.automl import SearchSpace, image_classification, image_object_detection
from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml.entities._inputs_outputs import Input
from azure.ai.ml.entities._job.automl.image import ImageClassificationSearchSpace, ImageObjectDetectionSearchSpace
from azure.ai.ml.entities._job.sweep.search_space import Choice, Uniform
from azure.ai.ml.exceptions import ValidationException


@pytest.mark.automl_test
@pytest.mark.parametrize(
    "task_name, expected_type",
    [(image_classification, ImageClassificationSearchSpace), (image_object_detection, ImageObjectDetectionSearchSpace)],
)
def test_searchspace(task_name, expected_type):
    settings = [
        Choice(["vitb16r224", "vits16r224"]),
        Uniform(0.001, 0.01),
        Choice([1, 5]),
    ]
    automl_job = task_name(
        training_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/train.csv"),
        target_column_name="label",
    )
    searchspace = SearchSpace(
        model_name=settings[0],
        learning_rate=settings[1],
        number_of_epochs=settings[2],
    )
    automl_job.search_space = [searchspace]
    assert automl_job.search_space[-1].model_name == settings[0]
    assert automl_job.search_space[-1].learning_rate == settings[1]
    assert automl_job.search_space[-1].number_of_epochs == settings[2]
    assert isinstance(automl_job.search_space[-1], expected_type)

    searchspace_dict = {
        "model_name": settings[0],
        "learning_rate": settings[1],
        "number_of_epochs": settings[2],
    }

    automl_job.search_space = [searchspace_dict]
    assert automl_job.search_space[-1].model_name == settings[0]
    assert automl_job.search_space[-1].learning_rate == settings[1]
    assert automl_job.search_space[-1].number_of_epochs == settings[2]
    assert isinstance(automl_job.search_space[-1], expected_type)

    automl_job.extend_search_space(searchspace)
    assert automl_job.search_space[-1].model_name == settings[0]
    assert automl_job.search_space[-1].learning_rate == settings[1]
    assert automl_job.search_space[-1].number_of_epochs == settings[2]
    assert isinstance(automl_job.search_space[-1], expected_type)

    assert len(automl_job.search_space) == 2


@pytest.mark.automl_test
@pytest.mark.parametrize(
    "task_name",
    [
        image_classification,
        image_object_detection,
    ],
)
def test_searchspace_invalid(task_name):
    automl_job = task_name(
        training_data=Input(type=AssetTypes.MLTABLE, path="https://foo/bar/train.csv"),
        target_column_name="label",
    )
    searchspace = SearchSpace(
        unspported_param=1,
    )
    with pytest.raises(ValidationException):
        automl_job.search_space = [searchspace]
    with pytest.raises(ValidationException):
        automl_job.search_space = [
            {
                "unspported_param": 1,
            }
        ]
    with pytest.raises(ValidationException):
        automl_job.extend_search_space([searchspace])
